import datetime
import os
import re
import subprocess

import lxml.etree as et
from lxml import objectify
import requests

from . import get_corpus_dir
from .transformations import (filename_to_doi, _get_base_page, LANDING_PAGE_SUFFIX,
                              URL_SUFFIX, plos_page_dict, doi_url, doi_to_url, doi_to_path)
from .plos_regex import validate_doi
from .elements import (parse_article_date, get_contrib_info,
                       Journal, License, match_contribs_to_dicts)
from .utils import dedent


class Article:
    """The primary object of a PLOS article, initialized by a valid PLOS DOI.

    """
    def __init__(self, doi, directory=None):
        """Creation of an article object.

        Usage:
        For the first time, you can use
        `article = Article(doi)`
        and then it and some attributes will be stored in memory.
        For creating articles after the first one, you can use:
        `article.doi = doi`
        This preserves the generic attributes and erases the article-specific ones
        (See also reset_memoized_attrs())
        Use this to more rapidly iterate through different articles.
        :param doi: The Digital Object Identifier of the article
        :type doi: str
        :param directory: where the local article XML file is located, defaults to None
        :type directory: str, optional
        """
        self.doi = doi
        self.directory = directory if directory else get_corpus_dir()
        self.reset_memoized_attrs()
        self._editor = None

    def __eq__(self, other):
        doi_eq = self.doi == other.doi
        dir_eq = self.directory == other.directory
        return doi_eq and dir_eq

    def __str__(self, exclude_refs=True):
        """Output when you print an article object on the command line.

        For parsing and viewing the XML of a local article. Should not be used for hashing
        Excludes <back> element (including references list) for easier viewing
        :param exclude_refs: remove references from the article tree (eases print viewing)
        """
        parser = et.XMLParser(remove_blank_text=True)
        tree = et.parse(self.filepath, parser)
        if exclude_refs:
            root = tree.getroot()
            back = tree.xpath('./back')
            if back:
                root.remove(back[0])
        local_xml = et.tostring(tree,
                                method='xml',
                                encoding='unicode',
                                pretty_print=True)
        return local_xml

    def __repr__(self):
        """Value of an article object when you call it directly on the command line.

        Shows the DOI and title of the article
        :returns: DOI and title
        :rtype: {str}
        """
        out = "DOI: {0}\nTitle: {1}".format(self.doi, self.title)
        return out


    def _repr_html_(self):
        """Nice display for Jupyter notebook"""

        titlestyle = 'display:inline-flex;'
        titletextstyle = 'margin-left:.5em;'
        titlelink = ('<span style="{titlestyle}"><a href="{url}">'
                     '<em>{title}</em></a></span>').format(
                        url=self.page,
                        title=self.title,
                        titlestyle=titlestyle+titletextstyle,
                    )

        doilink = '<span><a href="{url}"><code>{doi}</code></a></span>'.format(
                        url=self.doi_link(),
                        doi=self.doi,
                  )
        out = dedent("""<div>
        <span style="{titlestyle}">Title: {titlelink}</span></br>
        <span>DOI: <span>{doilink}
        </div>
        """).format(doilink=doilink, titlelink=titlelink, titlestyle=titlestyle)

        return out


    def reset_memoized_attrs(self):
        """Reset attributes to None when instantiating a new article object.

        For article attributes that are memoized and specific to that particular article
        (including the XML tree and whether the xml file is in the local directory),
        reset them when creating a new article object.
        """
        self._tree = None
        self._local = None
        self._contributors = None

    @property
    def doi(self):
        """The unique Digital Object Identifier for a PLOS article.

        See https://www.doi.org/
        :returns: DOI of the article object
        :rtype: {str}
        """
        return self._doi

    @property
    def text_viewer(self):
        """Command line application for viewing text to be used with
        open_in_viewer.

        Defaults to "open", which opens in whatever the default application is
        in your operating system for files ending in ".xml".

        Persists across article objects.
        Use with self.open_in_viewer() to open an article of interest.

        Check your text viewers documentation to learn how to launch it from the command line.
        For Sublime Text, see http://docs.sublimetext.info/en/latest/command_line/command_line.html
        :returns: command line shortcut for the text viewer
        :rtype: {str}
        """
        try:
            return self._text_viewer
        except AttributeError as e:
            print(("{}:\n"
                   "You need to assign a non-terminal text viewer "
                   "command able to be run on the CLI to self.text_viewer"
                   ).format(e))

    @text_viewer.setter
    def text_viewer(self, value="open"):
        """Sets the text viewer for all article objects.

        :param value: from self.text_viewer
        :type value: {str}
        """
        self._text_viewer = value

    @doi.setter
    def doi(self, d):
        """
        Using regular expressions, make sure the doi is valid before
        instantiating the article object.
        """
        if validate_doi(d) is False:
            raise Exception("Invalid format for PLOS DOI: {}".format(d))
        self.reset_memoized_attrs()
        self._doi = d


    def doi_link(self):
        """The link of the DOI, which redirects to the journal URL."""
        return doi_url + self.doi

    def get_remote_xml(self):
        """For an article, parse its XML file at the location of self.url.

        Uses the lxml element tree to create the string, which is saved to a local
        file when downloaded
        :returns: string of entire remote article file
        :rtype: {str}
        """
        remote_xml = et.tostring(self.remote_tree,
                                 method='xml',
                                 encoding='unicode')
        return remote_xml

    def open_in_viewer(self, text_viewer=None):
        """Open a local article file of interest in an external text viewer.

        :param text_viewer: set via self.text_viewer, defaults to None
        :type text_viewer: str, optional
        :raises: TypeError
        """
        if not (text_viewer or self.text_viewer):
            raise TypeError("You have not specified an text_viewer. Please do so.")

        subprocess.call([self._text_viewer, self.filepath])

    def open_in_browser(self):
        """Opens the landing page (HTML) of an article in default browser.

        This is also the URL that the DOI resolves to
        """
        subprocess.call(["open", self.page])

    def get_element_xpath(self, tag_path_elements=None, remote=False):
        """For a local article's root element, grab particular sub-elements via XPath location.

        Defaults to reading the element location for uncorrected proofs/versions of record
        The basis of every method and property looking for particular metadata fields
        :param article_root: the xml file for a single article
        :param tag_path_elements: xpath location in the XML tree of the article file
        :param remote: whether using the remote XML in self.remote_tree (defaults to False)
        :return: list of elements in the article with that xpath location
        """
        if tag_path_elements is None:
            tag_path_elements = ('/',
                                 'article',
                                 'front',
                                 'article-meta',
                                 'custom-meta-group',
                                 'custom-meta',
                                 'meta-value')
        tag_location = '/'.join(tag_path_elements)
        if remote:
            root = self.remote_tree.getroot()
        else:
            root = self.root
        return root.xpath(tag_location)

    def get_dates(self, string_=False, string_format='%Y-%m-%d'):
        """For an individual article, get all of its dates, including publication date (pubdate), submission date.

        Defaults to datetime objects
        :param string_: whether to return dates as a dictionary of strings
        :param string_format: if string_ is True, the format to return the dates in
        :return: dict of date types mapped to datetime objects for that article
        :rtype: {dict}
        """
        dates = {}
        # first location is where pubdate and date added to collection are
        tag_path_1 = ["/",
                      "article",
                      "front",
                      "article-meta",
                      "pub-date"]
        element_list_1 = self.get_element_xpath(tag_path_elements=tag_path_1)
        for element in element_list_1:
            pub_type = element.get('pub-type')
            try:
                date = parse_article_date(element)
            except ValueError:
                print('Error getting pubdates for {}'.format(self.doi))
                date = ''
            dates[pub_type] = date

        # second location is where historical dates are, including submission and acceptance
        tag_path_2 = ["/",
                      "article",
                      "front",
                      "article-meta",
                      "history"]
        element_list_2 = self.get_element_xpath(tag_path_elements=tag_path_2)
        for element in element_list_2:
            for part in element:
                date_type = part.get('date-type')
                try:
                    date = parse_article_date(part)
                except ValueError:
                    print('Error getting history dates for {}'.format(self.doi))
                    date = ''
                dates[date_type] = date

        # third location is for vor updates when it's updated (see `proof(self)`)
        rev_date = ''
        if self.proof == 'vor_update':
            tag_path = ('/',
                        'article',
                        'front',
                        'article-meta',
                        'custom-meta-group',
                        'custom-meta')
            xpath_results = self.get_element_xpath(tag_path_elements=tag_path)
            for result in xpath_results:
                if result.xpath('./meta-name')[0].text == 'Publication Update':
                    rev_date_string = result.xpath('./meta-value')[0].text
                    rev_date = datetime.datetime.strptime(rev_date_string, '%Y-%m-%d')
                    break
                else:
                    pass
        dates['updated'] = rev_date

        if string_:
            # can return dates as strings instead of datetime objects if desired
            for key, value in dates.items():
                if value:
                    dates[key] = value.strftime(string_format)

        return dates

    def dates_debug(self):
        """Whether the dates in self.get_dates() are in the correct order.

        check whether date received is before date accepted, is before pubdate
        accounts for potentially missing date fields
        :return: if dates are in right order or not
        :rtype: bool
        """
        dates = self.get_dates()
        if dates.get('received', '') and dates.get('accepted', ''):
            if dates['received'] <= dates['accepted'] <= dates['epub']:
                order_correct = True
            else:
                order_correct = False
        elif dates.get('received', ''):
            if dates['received'] <= dates['epub']:
                order_correct = True
            else:
                order_correct = False
        elif dates.get('accepted', ''):
            if dates['accepted'] <= dates['epub']:
                order_correct = True
            else:
                order_correct = False
        else:
            order_correct = True

        return order_correct

    @property
    def volume(self):
        """Volume of the article."""
        return int(self.root.xpath('/article/front/article-meta/volume')[0].text)

    @property
    def issue(self):
        """Issue of the article."""
        return int(self.root.xpath('/article/front/article-meta/issue')[0].text)

    @property
    def elocation(self):
        """Elocation ID of the article."""
        return self.root.xpath('/article/front/article-meta/elocation-id')[0].text

    def get_aff_dict(self):
        """For a given PLOS article, get list of contributor-affiliated institutions.

        Uses "rid"s to map individual contributors to their institutions
        More about rids: https://jats.nlm.nih.gov/archiving/tag-library/1.1/attribute/rid.html
        See also get_rid_dict()
        :returns: Dictionary of footnote ids to institution information
        :rtype: {dict}
        """
        tags_to_aff = ["/",
                       "article",
                       "front",
                       "article-meta"]
        article_aff_elements = self.get_element_xpath(tag_path_elements=tags_to_aff)
        aff_dict = {}
        aff_elements = [el
                        for aff_element in article_aff_elements
                        for el in aff_element.getchildren()
                        ]
        for el in aff_elements:
            if el.tag == 'aff':
                if el.getchildren():
                    for sub_el in el.getchildren():
                        if sub_el.tag == 'addr-line':
                            try:
                                aff_text_fixed = ' '.join([aff_string.strip() for aff_string in sub_el.text.splitlines()])
                            except AttributeError:
                                aff_text_fixed = et.tostring(sub_el, encoding='unicode', method='text')
                            aff_dict[el.attrib['id']] = aff_text_fixed
                else:
                    # the address for some affiliations is not wrapped in an addr-line tag
                    aff_dict[el.attrib['id']] = el.text.replace('\n', '').replace('\r', '').replace('\t', '')
        return aff_dict

    def get_fn_dict(self):
        """For a given PLOS article, get list of footnotes.

        Used with rids to map individual contributors to their institutions
        More about rids: https://jats.nlm.nih.gov/archiving/tag-library/1.1/attribute/rid.html
        See also get_rid_dict()
        :returns: Dictionary of footnote ids to institution information
        :rtype: {dict}
        """
        tags_to_fn = ["/",
                      "article",
                      "front",
                      "article-meta",
                      "author-notes"]
        article_fn_elements = self.get_element_xpath(tag_path_elements=tags_to_fn)
        fn_dict = {}
        fn_elements = [el
                       for fn_element in article_fn_elements
                       for el in fn_element.getchildren()
                       ]
        for el in fn_elements:
            if el.attrib.get('id'):
                if el.getchildren():
                    for sub_el in el.getchildren():
                        if sub_el.tag == 'email':
                            pass
                        else:
                            fn_dict[el.attrib['id']] = sub_el.text
                else:
                    # in case is at top-level of element
                    fn_dict[el.attrib['id']] = el.text.replace('\n', '').replace('\r', '').replace('\t', '')
        return fn_dict

    def get_corr_author_emails(self):
        """For an article, grab the email addresses of the corresponding authors.
        Parses the list of emails and groups by rid or by initials, if present.
        Can handle multiple emails for multiple authors if formatted correctly.
        The email addresses are in an element of author notes. While most articles have one corresponding
        author with one email address, sometimes there are 1) multiple authors, and/or 2) multiple emails per
        author. In the first case, author initials are used in the text to separate emails. In the second case,
        a comma is used to separate emails. Initials are how emails can be matched to multiple
        authors. See also `match_author_names_to_emails()` for the back-up method of name matching.
        :return: dictionary of rid or author initials mapped to list of email address(es)
        :rtype: {dict}
        """
        tag_path = ["/",
                    "article",
                    "front",
                    "article-meta",
                    "author-notes"]
        try:
            author_notes_element = self.get_element_xpath(tag_path_elements=tag_path)[0]
        except IndexError:
            # no emails found
            return {}
        corr_emails = {}
        email_list = []
        for note in author_notes_element:
            if note.tag == 'corresp':
                author_info = note.getchildren()
                for i, item in enumerate(author_info):
                    # if author initials are in the same field as email address
                    if item.tag == 'email' and item.text and all(x in item.text for x in ('(', ')')):
                        email_info = item.text.split(' ')
                        for i, info in enumerate(email_info):
                            # prune out non-letters from initials & email
                            email_info[i] = re.sub(r'[^a-zA-Z0-9=@\.+-]', '', info)
                        try:
                            corr_emails[email_info[1]] = [email_info[0]]
                        except IndexError:
                            print('Error parsing emails for {}'.format(self.doi))
                            pass

                    # if no author initials (one corr author)
                    elif item.tag == 'email' and item.tail is None and item.text:
                        email_list.append(item.text)
                        if item.text == '':
                            print('No email available for {}'.format(self.doi))
                        if note.attrib['id']:
                            corr_emails[note.attrib['id']] = email_list
                        else:
                            corr_emails['cor001'] = email_list

                    # if more than one email per author; making sure no initials present (comma ok)
                    elif item.tag == 'email' and re.sub(r'[^a-zA-Z0-9=]', '', str(item.tail)) is None:
                        try:
                            if author_info[i+1].tail is None:
                                email_list.append(item.text)
                            elif author_info[i+1].tail:
                                corr_initials = re.sub(r'[^a-zA-Z0-9=]', '', author_info[i+1].tail)
                                if not corr_emails.get(corr_initials):
                                    corr_emails[corr_initials] = [item.text]
                                else:
                                    corr_emails[corr_initials].append(item.text)
                        except IndexError:
                            email_list.append(item.text)
                            corr_emails[note.attrib['id']] = email_list
                            if i > 1:
                                print('Error handling multiple email addresses for {} in {}'
                                      .format(et.tostring(item), self.doi))
                        if item.text == '':
                            print('No email available for {}'.format(self.doi))

                    # if author initials included (more than one corr author)
                    elif item.tag == 'email' and item.tail:
                        corr_email = item.text
                        corr_initials = re.sub(r'[^a-zA-Z0-9=]', '', item.tail)
                        if not corr_initials:
                            try:
                                corr_initials = re.sub(r'[^a-zA-Z0-9=]', '', author_info[i+1].tail)
                            except (IndexError, TypeError) as e:
                                corr_initials = note.attrib['id']
                                if not corr_initials:
                                    print('email parsing is weird for', self.doi)
                        if not corr_emails.get(corr_initials):
                            corr_emails[corr_initials] = [corr_email]
                        else:
                            corr_emails[corr_initials].append(corr_email)
                    else:
                        pass
        if not corr_emails:
            author_notes_field = et.tostring(author_notes_element, method='text', encoding='unicode')
            if '@' in author_notes_field:
                regex_email = r'[\w\.-]+@[\w\.-]+'
                email_finder = re.compile(regex_email)
                email_list = email_finder.findall(author_notes_field)
                if email_list:
                    corr_emails['cor001'] = email_list
        return corr_emails

    def get_contributions_dict(self):
        """For articles that don't use the CREDiT taxonomy, compile a dictionary of author
        contribution types matched to author initials.
        Work in progress!!
        Works for highly formatted lists with subelements (e.g. '10.1371/journal.pone.0170354') and structured single strings
        (e.g. '10.1371/journal.pone.0050782'), but still fails for unusual strings (e.g, '10.1371/journal.pntd.0000072')
        See also get_credit_taxonomy() for the CREDiT taxonomy version.
        TODO: Use regex to properly separate author roles from initials for unusual strings.
        :return: dictionary mapping author initials to their author contributions/roles.
        """
        if self.type_ in ['correction', 'retraction', 'expression-of-concern']:
            # these article types don't have proper 'authors'
            return {}
        tag_path = ["/",
                    "article",
                    "front",
                    "article-meta",
                    "author-notes"]
        try:
            author_notes_element = self.get_element_xpath(tag_path_elements=tag_path)[0]
        except IndexError:
            return {}
        author_contributions = {}
        contrib_dict = {}
        initials_list = []
        for note in author_notes_element:
            if note.attrib.get('fn-type', None) == 'con':
                try:
                    # for highly structured lists with sub-elements for each item
                    # Example: 10.1371/journal.pone.0170354'
                    con_element = note[0][0]
                    con_list = con_element.getchildren()
                    for con_item in con_list:
                        try:
                            contribution = con_item[0][0].text.rstrip(':')
                            contributor_initials = (con_item[0][0].tail.lstrip(' ').rstrip('.')).split(' ')
                            initials_list.extend(contributor_initials)
                            contrib_dict[contribution] = contributor_initials
                        except (IndexError, AttributeError) as e:
                            print('Error parsing contributions item {}: {}'.format(self.doi, et.tostring(con_item,
                                                                                                         encoding='unicode',
                                                                                                         method='xml')))
                            pass
                except IndexError:
                    # for single strings, though it doesn't parse all of them correctly.
                    # Example: '10.1371/journal.pone.0050782'
                    contributions = note[0].text
                    if contributions is None:
                        print('Error parsing contributions for {}: {}'.format(self.doi, et.tostring(con_element,
                                                                                                    encoding='unicode',
                                                                                                    method='xml')))
                        return {}
                    contribution_list = re.split(': |\. ', contributions)
                    contribb_dict = dict(list(zip(contribution_list[::2], contribution_list[1::2])))
                    for k, v in contribb_dict.items():
                        v_new = v.split(' ')
                        v_new = [v.rstrip('.').strip('\n') for v in v_new]
                        contrib_dict[k.strip('\n')] = v_new
                        initials_list.extend(v_new)

        for initials in (set(initials_list)):
            contrib_list = []
            for k, v in contrib_dict.items():
                if initials in v:
                    contrib_list.append(k)
            author_contributions[initials] = contrib_list
        return author_contributions

    def get_contributors_info(self):
        """Get and organize information about each contributor for an article.
        This includes both authors and editors of the article.
        This function both creates article-level dictionaries of contributor information,
        as well as parses individual <contrib> elements. It reconciles the dicts together
        using a number of external functions from article_elements.py
        :returns: dictionary of metadata for each <contrib> element
        :rtype: list of dicts
        """

        # TODO: param to remove unnecessary fields (initials) and dicts (rid_dict)
        # TODO: also get funding information, data availability, COI, etc

        # get dictionary of ids to institutional affiliations & all other footnotes
        aff_dict = self.get_aff_dict()
        fn_dict = self.get_fn_dict()
        aff_dict.update(fn_dict)
        matching_error = False

        # get dictionary of corresponding author email addresses
        email_dict = self.get_corr_author_emails()

        # get author contributions (if no credit taxonomy)
        credit_dict = self.get_contributions_dict()

        # get list of contributor elements (one per contributor)
        tag_path = ["/",
                    "article",
                    "front",
                    "article-meta",
                    "contrib-group",
                    "contrib"]
        contrib_list = self.get_element_xpath(tag_path_elements=tag_path)
        contrib_dict_list = []

        error_printed = False

        # iterate through each contributor
        for contrib in contrib_list:
            # initialize contrib dict with default fields
            contrib_keys = ['contrib_initials',
                            'given_names',
                            'surname',
                            'group_name',
                            'ids',
                            'rid_dict',
                            'contrib_type',
                            'author_type',
                            'editor_type',
                            'email',
                            'affiliations',
                            'author_roles',
                            'footnotes'
                            ]
            contrib_dict = dict.fromkeys(contrib_keys, None)
            try:
                contrib_dict.update(get_contrib_info(contrib))
            except TypeError:
                # minimize number of times this prints out
                if not error_printed:
                    print('Error getting contrib info for {}'.format(self.doi, self.type_))
                    error_printed = True
                else:
                    pass

            # map affiliation footnote ids to the actual institutions
            try:
                if contrib_dict.get('rid_dict', None) is not None:
                    contrib_dict['affiliations'] = [aff_dict.get(aff, "")
                                                    for k, v in contrib_dict['rid_dict'].items()
                                                    for aff in v
                                                    if k == 'aff'
                                                    ]
            except (TypeError, AttributeError) as e:
                print('error constructing affiliations for {}: {} {}'
                      .format(self.doi,
                              contrib_dict.get('given_names'),
                              contrib_dict.get('surname')))
                contrib_dict['affiliations'] = [""]
            try:
                contrib_dict['footnotes'] = [aff_dict.get(aff, "")
                                             for k, v in contrib_dict['rid_dict'].items()
                                             for aff in v
                                             if k == 'fn'
                                             ]
            except AttributeError:
                print('error constructing footnote matches for {}: {} {}'
                      .format(self.doi,
                              contrib_dict.get('given_names'),
                              contrib_dict.get('surname')))
                contrib_dict['affiliations'] = [""]
            # make list of all contribs
            contrib_dict_list.append(contrib_dict)

        # match authors to credit_dicts (from author notes) if necessary
        if credit_dict:
            author_list = [author for author in contrib_dict_list
                           if author.get('contrib_type', None) == 'author']
            author_list, credit_matching_error = match_contribs_to_dicts(author_list,
                                                                         credit_dict,
                                                                         contrib_key='author_roles')
            for author in author_list:
                role_list = author.get('author_roles', None)
                author['author_roles'] = {'author_notes': role_list}

            if credit_matching_error:
                print('Warning: authors not matched correctly to author_roles for {}'
                      .format(self.doi))

        # match corresponding authors to email addresses
        corr_author_list = [contrib for contrib in contrib_dict_list if contrib.get('author_type', None) == 'corresponding']
        if not corr_author_list and email_dict:
            print('Email but no corresponding author found for {}'.format(self.doi))
            # matching_error = True
        if corr_author_list and not email_dict:
            print('Corr emails not found for {}'.format(self.doi))
            matching_error = True
        if len(corr_author_list) == 1:
            corr_author = corr_author_list[0]
            try:
                corr_author['email'] = email_dict[corr_author['rid_dict']['corresp'][0]]
            except KeyError:
                if len(email_dict) == 1:
                    corr_author['email'] = list(email_dict.values())[0]
                else:
                    print('one_corr_author error finding email for {} in {}'.format(corr_author, email_dict))
                    matching_error = True
        elif email_dict and len(corr_author_list) > 1 and len(set([tuple(x) for x in email_dict.values()])) > 1:
            corr_author_list, matching_error = match_contribs_to_dicts(corr_author_list,
                                                                       email_dict,
                                                                       contrib_key='email')
        elif len(corr_author_list) > 1:
            if email_dict and (len(email_dict) == 1 or len(set([tuple(x) for x in email_dict.values()])) == 1):
                # if there's only one email address, use it for all corr authors
                for corr_author in corr_author_list:
                    corr_author['email'] = list(email_dict.values())[0]
            else:
                matching_error = True
        else:
            corr_author_list = []

        match_error_printed = False
        if email_dict and len(email_dict) > len(corr_author_list) > 0:
                print('Contributing author email included for {}'
                      .format(self.doi))
                match_error_printed = True
        elif email_dict and 1 < len(email_dict) < len(corr_author_list):
            print('{} corresponding author email(s) missing for {}'
                  .format(len(corr_author_list) - len(email_dict), self.doi))
            match_error_printed = True

        if matching_error and email_dict and not match_error_printed:
            print('Warning: corresponding authors not matched correctly to email addresses for {}'
                  .format(self.doi))
        return contrib_dict_list

    def get_related_dois(self):
        """For a given article, get the list of DOIs of related PLOS articles.
        Creates a dictionary of related dois & their type from the <related-articles> xpath location
        Use primarily to map amendment notifications to articles that have been amended
        :return: dictionary of related DOIs
        :rtype: dict
        """
        related_article_elements = self.get_element_xpath(tag_path_elements=["/",
                                                                             "article",
                                                                             "front",
                                                                             "article-meta",
                                                                             "related-article"])
        related_article_dict = {}

        if related_article_elements:
            for elem in related_article_elements:
                related_doi = elem.attrib
                related_article = related_doi['{http://www.w3.org/1999/xlink}href']
                related_article = related_article.lstrip('info:doi/')
                if not related_article_dict.get(elem.attrib['related-article-type'], None):
                    # begin building the list of DOIs with that related-article-type
                    related_article_dict[elem.attrib['related-article-type']] = [related_article]
                else:
                    # there is more than one article with the same related-article-type
                    related_article_dict[elem.attrib['related-article-type']].append(related_article)
        else:
            # no related articles exist
            pass
        return related_article_dict

    def check_if_link_works(self):
        """See if a link is valid (i.e., returns a '200' to the HTML request).

        Used for checking a URL to a PLOS article's landing page or XML file on journals.plos.org
        Full list of potential status codes: https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
        :return: boolean if HTTP status code returned available or unavailable,
        "error" if a different status code is returned than 200 or 404
        """
        request = requests.get(self.url)
        if request.status_code == 200:
            return True
        elif request.status_code == 404:
            return False
        else:
            return 'error'

    def check_if_doi_resolves(self, plos_valid=True):
        """Whether a PLOS DOI resolves via dx.doi.org to the correct article landing page.

        If the link works, make sure that it points to the same DOI
        Checks first if it's a valid DOI or see if it's a redirect.
        :return: 'works' if works as expected, 'doesn't work' if it doesn't resolve correctly,
        or if the metadata DOI doesn't match self.doi, return the metadata DOI
        """
        if plos_valid and validate_doi(self.doi) is False:
            return "Not valid PLOS DOI structure"
        url = "http://dx.doi.org/" + self.doi
        if self.check_if_link_works() is True:
            headers = {"accept": "application/vnd.citationstyles.csl+json"}
            r = requests.get(url, headers=headers)
            r_doi = r.json()['DOI']
            if r_doi == self.doi:
                return "works"
            else:
                return r_doi
        else:
            return "doesn't work"

    @property
    def xml(self):
        """Returns string from local xml file.
        """
        if self.tree is None:
            return None
        else:
            local_xml = et.tostring(self.tree,
                                    method='xml',
                                    encoding='unicode')
            return local_xml

    @property
    def tree(self):
        """The element tree object created from an article's local XML file

        See http://lxml.de/api/lxml.etree._ElementTree-class.html
        After accessing tree for the first time, it stores as an attribute
        :returns: article's element tree
        :rtype: {lxml.etree._ElementTree-class} or None
        """
        if self._tree is None:
            if self.local:
                local_element_tree = et.parse(self.filepath)
                self._tree = local_element_tree
            else:
                print("Local article file not found: {}".format(self.filepath))
                return None
        else:
            pass
        return self._tree

    @property
    def root(self):
        """Get the root (base) element of an article.
        """
        return self.tree.getroot()

    def get_page(self, page_type='article'):
        """Get any of the PLOS URLs associated with a particular DOI.

        Based on `get_page_base()`, which customizes the beginning URL by journal.
        :param page_type: one of the keys in `plos_page_dict`, defaults to article
        """
        BASE_LANDING_PAGE = _get_base_page(self.journal)
        try:
            page = BASE_LANDING_PAGE + LANDING_PAGE_SUFFIX.format(plos_page_dict[page_type],
                                                                  self.doi)
            if page_type == 'assetXMLFile':
                page += URL_SUFFIX
        except KeyError:
            raise Exception('Invalid page_type; value must be one of the following: {}'.format(list(plos_page_dict.keys())))
        return page

    @property
    def page(self):
        """ The URL of the landing page for an article.

        Where to access an article's HTML version
        """
        return self.get_page()

    @property
    def url(self):
        """The direct url of an article's XML file.
        """
        return self.get_page(page_type='assetXMLFile')

    @property
    def taxonomy(self):
        """Taxonomy information. For a complete list of subject areas see
        https://github.com/PLOS/plos-thesaurus
        """
        tag_path_elements = ('/',
                             'article',
                             'front',
                             'article-meta',
                             'article-categories')
        e_list = self.get_element_xpath(tag_path_elements=tag_path_elements)
        subjs_dict = {}
        for subj in e_list[0].getchildren():
            try:
                sbjindex = subj.values()[0].strip()
                if sbjindex in subjs_dict:
                    subjs_dict[sbjindex].append(tuple(e.text for e in subj.iter('subject')))
                else:
                    subjs_dict[sbjindex] = [tuple(e.text for e in subj.iter('subject'))]
            except IndexError:
                if 'No subject' in subjs_dict:
                    subjs_dict['No subject'].append(tuple(e.text for e in
                                                 subj.iter('subject')))
                else:
                    subjs_dict['No subject'] = [tuple(e.text for e in
                                                 subj.iter('subject'))]
        return subjs_dict

    @property
    def filepath(self):
        """The path on the local file system to a given article's XML file.
        """
        if 'annotation' in self.doi:
            article_path = os.path.join(self.directory, 'plos.correction.' + self.doi.split('/')[-1] + '.xml')
        else:
            article_path = os.path.join(self.directory, self.doi.lstrip('10.1371/') + '.xml')
        return article_path

    @property
    def filename(self):
        """The basename of the article's XML file.
        """
        return os.path.basename(self.filepath)

    @property
    def local(self):
        """Boolean of whether the article is stored locally or not.

        Stored as attribute after first access
        """
        if self._local is None:
            self._local = os.path.isfile(self.filepath)
        else:
            pass
        return self._local

    @property
    def proof(self):
        """
        For a single article in a directory, check whether it is an 'uncorrected proof' or a
        'VOR update' to the uncorrected proof, or neither.
        :return: proof status if it exists
        :rtype: str
        """
        xpath_results = self.get_element_xpath()
        proof = ''
        for result in xpath_results:
            if result.text == 'uncorrected-proof':
                proof = 'uncorrected_proof'
            elif result.text == 'vor-update-to-uncorrected-proof':
                proof = 'vor_update'
        return proof

    @property
    def remote_proof(self):
        """
        For a single article online, check whether it is an 'uncorrected proof' or a
        'VOR update' to the uncorrected proof, or neither.
        :return: proof status if it exists; otherwise, None
        """
        xpath_results = self.get_element_xpath(remote=True)
        proof = ''
        for result in xpath_results:
            if result.text == 'uncorrected-proof':
                proof = 'uncorrected_proof'
            elif result.text == 'vor-update-to-uncorrected-proof':
                proof = 'vor_update'
        return proof

    @property
    def remote_tree(self):
        """Gets the lxml element tree of an article from its remote URL.

        Can compare local (self.xml) to remote versions of XML
        :returns: article's online element tree
        :rtype: {lxml.etree._ElementTree-class}
        """
        return et.fromstring(requests.get(self.url).content)

    @property
    def journal(self):
        """Journal that an article was published in.
        Can be PLOS Biology, Medicine, Neglected Tropical Diseases, Pathogens,
        Genetics, Computational Biology, ONE, or the now defunct Clinical Trials.
        Relies on a simple doi_to_journal transform when possible, and uses `Journal().parse_plos_journal()`
        for the "annotation" DOIs that don't have that journal information in the DOI.
        """
        if 'annotation' not in self.doi:
            journal = Journal.doi_to_journal(self.doi)
        else:
            journal_meta = self.root.xpath('/article/front/journal-meta')[0]
            journal = str(Journal(journal_meta))
        return journal

    @property
    def title(self):
        """For an individual PLOS article, get its title.

        :return: string of article title at specified xpath location
        """
        title = self.get_element_xpath(tag_path_elements=["/",
                                                          "article",
                                                          "front",
                                                          "article-meta",
                                                          "title-group",
                                                          "article-title"])
        title_text = et.tostring(title[0], encoding='unicode', method='text', pretty_print=True)
        title_cleaned = " ".join(title_text.split())
        return title_cleaned

    @property
    def rich_title(self):
        """For an individual PLOS article, get its title with HTML formatting.

        Preserves HTML formatting but removes all other XML tagging, namespace/xlink info, etc.
        Doesn't do xpath directly on `self.root` so can deannotate separate object
        See http://lxml.de/objectify.html#how-data-types-are-matched for more info on deannotate process
        Exceptions that still need handling:
        10.1371/journal.pone.0179720, 10.1371/journal.pone.0068479, 10.1371/journal.pone.0069681,
        10.1371/journal.pone.0068965, 10.1371/journal.pone.0083868, 10.1371/journal.pone.0069554,
        10.1371/journal.pone.0068324, 10.1371/journal.pone.0067986, 10.1371/journal.pone.0068704,
        10.1371/journal.pone.0068492, 10.1371/journal.pone.0068764, 10.1371/journal.pone.0068979,
        10.1371/journal.pone.0068544, 10.1371/journal.pone.0069084, 10.1371/journal.pone.0069675

        :return: string of article title at specified xpath location
        """
        root = self.root
        objectify.deannotate(root, cleanup_namespaces=True, xsi_nil=True)
        art_title = root.xpath("/article/front/article-meta/title-group/article-title")
        art_title = art_title[0]
        try:
            text = art_title.text
            if text is None:
                text = ''
            text += ''.join(et.tostring(child, encoding='unicode') if child.tag not in ('ext-link', 'named-content', 'sc', 'monospace') \
                                                                   else child.text + child.tail if child.tail is not None \
                                                                   else child.text
                            for child in art_title.getchildren())
            title = text.replace(' xmlns:xlink="http://www.w3.org/1999/xlink"', '') \
                        .replace(' xmlns:mml="http://www.w3.org/1998/Math/MathML"', '') \
                        .replace(' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance', '')
        except TypeError:
            # try to rewrite so this isn't needed
            print('Error processing article title for {}'.format(self.doi))
            title = et.tostring(art_title, method='text', encoding='unicode')
        return title

    @property
    def pubdate(self):
        """The date an article was published online.

        :returns: article publication date
        :rtype: {datetime.datetime}
        """
        dates = self.get_dates()
        return dates['epub']

    @property
    def revdate(self):
        """The date an article's version-of-record (`proof(self)` == 'vor_update') was published online.

        :returns: article revision date
        :rtype: {datetime.datetime}
        """
        dates = self.get_dates()
        return dates['updated']

    @property
    def license(self):
        """Return dictionary of CC license information from the license field."""
        permissions = self.root.xpath('/article/front/article-meta/permissions')[0]
        return dict(License(permissions, self.doi))

    @property
    def contributors(self):
        """ List of contributors to an article.

        Including authors and editors
        Stores as attribute after first access
        :returns: list of dictionaries for each contributor
        :rtype: {list}
        """
        if self._contributors is None:
            self._contributors = self.get_contributors_info()
        else:
            pass
        return self._contributors

    @property
    def authors(self):
        """List of authors of an article. Including contributing and corresponding.

        For more about authorship criteria, see https://journals.plos.org/plosone/s/authorship
        """
        contributors = self.contributors
        return [contrib for contrib in contributors if contrib.get('contrib_type', None) == 'author']

    @property
    def corr_author(self):
        """List of corresponding authors of an article.
        """
        contributors = self.contributors
        return [contrib for contrib in contributors if contrib.get('author_type', None) == 'corresponding']

    @property
    def editor(self):
        """The editor on the article.

        For more about the editorial process, see https://journals.plos.org/plosone/s/editorial-and-peer-review-process
        """
        contributors = self.contributors
        return [contrib for contrib in contributors if contrib.get('contrib_type', None) == 'editor']

    @property
    def emails(self):
        """List of emails of corresponding author(s).
        Unlike get_corr_author_emails() dict, it does not differentiate by author.
        Joins multiple emails into a single list.
        :return: list of corresponding author email addresses
        """
        email_dict = self.get_corr_author_emails()
        email_list = []
        for k, v in email_dict.items():
            email_list.extend(v)
        return email_list

    def emails_to_string(self):
        """Produces string of emails of corresponding author(s).
        Joins multiple emails into a single string, separated by semi-colons.
        Used for exporting to .csv
        :return: string of corresponding author email addresses
        """
        return '; '.join(self.emails)

    @property
    def type_(self):
        """For an article file, get its JATS article type.

        Used primarily to find Correction (and thereby corrected) articles
        :return: JATS article_type at that xpath location
        """
        type_element_list = self.get_element_xpath(tag_path_elements=["/",
                                                                      "article"])
        return type_element_list[0].attrib['article-type']

    @property
    def plostype(self):
        """For an article file, get its PLOS article type.

        This format is less standardized than the JATS article type (self.type_)
        :return: PLOS article_type at that xpath location
        """
        article_categories = self.get_element_xpath(tag_path_elements=["/",
                                                                       "article",
                                                                       "front",
                                                                       "article-meta",
                                                                       "article-categories"])
        subject_list = article_categories[0].getchildren()

        for i, subject in enumerate(subject_list):
            if subject.get('subj-group-type') == "heading":
                subject_instance = subject_list[i][0]
                s = ''
                for text in subject_instance.itertext():
                    s = s + text
                    plos_article_type = s
        return plos_article_type

    @property
    def dtd(self):
        """Document Type Definition for an article.
        For more information on these DTD tagsets, see https://jats.nlm.nih.gov/1.1d3/ and https://dtd.nlm.nih.gov/3.0/
        """
        dtd = self.get_element_xpath(tag_path_elements=["/",
                                                        "article"])
        try:
            dtd = dtd[0].attrib['dtd-version']
            if str(dtd) == '3.0':
                dtd = 'NLM 3.0'
            elif dtd == '1.1d3':
                dtd = 'JATS 1.1d3'
        except KeyError:
            print('Error parsing DTD from', self.doi)
            dtd = 'N/A'
        return dtd

    @property
    def abstract(self):
        """For an individual PLOS article, get the string of the abstract content.

        PLOS articles can have multiple abstract fields at the same XPath location,
        however the actual abstract is distinguished by having no attributes (`[count(@*)=0]`).
        Info about the article abstract: https://journals.plos.org/plosone/s/submission-guidelines#loc-abstract
        :return: plain-text string of content in abstract
        """
        abstract_list = self.get_element_xpath(tag_path_elements=["/",
                                                                  "article",
                                                                  "front",
                                                                  "article-meta",
                                                                  "abstract[count(@*)=0]"])
        if abstract_list:
                abstract = abstract_list[0]
                assert len(abstract_list) == 1

                abstract_text = et.tostring(abstract[0], encoding='unicode', method='text')
        else:
            if self.type_ == 'research-article' and self.plostype == 'Research Article':
                print('No abstract found for research article {}'.format(self.doi))

            abstract_text = ''

        # clean up text: rem white space, new line marks, blank lines
        abstract_text = abstract_text.strip().replace('  ', '')
        abstract_text = os.linesep.join([s for s in abstract_text.splitlines() if s])

        return abstract_text

    @property
    def body(self):
        """
        For an individual PLOS article, get the string of the body content.

        :returns: main body of the article
        :rtype: {str}
        """

        xml_tree = et.parse(self.filepath)
        root = xml_tree.getroot()

        # limit the text to the body section
        body = root.find('./body')

        # remove supplementary material section
        for sec in body.findall('.//sec'):
            if 'supplementary-material' in [element.tag for element in sec]:
                parent=sec.getparent()
                parent.remove(sec)

        # remove unwanted xml elements
        remove_list = ['tr','td','caption','fig','graphic','table-wrap','ext-link']
        for tag in remove_list:
            for element in body.findall('.//'+tag):
                parent=element.getparent()
                parent.remove(element)

        # convert XML to string
        body_text = et.tostring(body, encoding='utf8', method='text').decode()

        return(body_text)

    @property
    def amendment(self):
        """Whether the JATS article type is a correction, retraction, or expression of concern.

        These are the three article types ('amendments') that potentially warrant a change in the original article
        that they reference (i.e., the 'related-doi'.)
        See https://jats.nlm.nih.gov/archiving/tag-library/1.1/attribute/article-type.html
        :returns: True if an amendment article type, False if not
        :rtype: {bool}
        """
        if self.type_ in ['correction', 'retraction', 'expression-of-concern']:
            return True
        else:
            return False

    @property
    def related_dois(self):
        """PLOS DOIs related to current article.

        Compresses the values of `self.get_related_dois()` dictionary into a single list of DOI strings
        More strict for which keys to include for corrections, retractions, and expressions of concern, the three
        amendment article types.
        :returns: list of related DOIs
        :rtype: list
        """
        doi_list = []
        related_doi_dict = self.get_related_dois()
        if self.amendment:
            # only use certain keys if an amendment article
            if self.type_ == 'correction':
                attrib_name = 'corrected-article'
            elif self.type_ == 'retraction':
                attrib_name = 'retracted-article'
            elif self.type_ == 'expression-of-concern':
                attrib_name = 'object-of-concern'
            for k, v in related_doi_dict.items():
                if k == attrib_name:
                    doi_list = v
                    break
            if not doi_list:
                doi_list = [related for related_list in related_doi_dict.values() for related in related_list]
                print('{} has incorrect related_doi field attribute'.format(self.doi))

        else:
            # flatten all dict values if not an amendment article
            if related_doi_dict:
                for k, v in related_doi_dict.items():
                    doi_list.extend(v)

        return doi_list

    @property
    def correction(self):
        """Get the DOIs of all corrections type articles that correct the current article.

        Some PLOS articles include a 'correction-forward' related-article-type, meaning
        an article that has been issued a correction is linked to its correcting article(s).
        Only for the SIX PLOS journals (i.e. not on PLOS ONE).
        Usually there is only one DOI, unless the article has been issued multiple corrections.
        :return: DOIs of the correction articles
        :rtype: list
        """
        correction_doi = ''
        related_dois = self.get_related_dois()
        for k, v in related_dois.items():
            if k == 'correction-forward':
                correction_doi = v
                break
        return correction_doi

    @property
    def counts(self):
        """For a single article, return a dictionary of the several counts functions that are available.

        Dictionary format for XML tags: {figures: fig-count, pages: page-count, tables: table-count}
        For articles without the figure and table counts fields, calculates those values using XPath.
        :return: counts dictionary of number of figures, pages, and tables in the article
        """
        counts = {}

        tag_path = ["/",
                    "article",
                    "front",
                    "article-meta",
                    "counts"]
        count_element_list = self.get_element_xpath(tag_path_elements=tag_path)
        for count_element in count_element_list:
            for count_item in count_element:
                count = count_item.get('count')
                count_type = count_item.tag
                counts[count_type] = int(count)
        if len(counts) > 3:  # this shouldn't happen
            print(counts)
        if 'fig-count' not in counts:
            counts['fig-count'] = len(self.root.xpath('.//fig'))
        if 'table-count' not in counts:
            counts['table-count'] = len(self.root.xpath('.//table-wrap'))
        return counts

    @property
    def word_count(self):
        """For an article, get how many words are in the body.

        :return: count of words in the body of the PLOS article
        """
        body_element = self.get_element_xpath(tag_path_elements=["/",
                                                                 "article",
                                                                 "body"])
        try:
            body_text = et.tostring(body_element[0], encoding='unicode', method='text')
            body_word_count = len(body_text.split(" "))
        except IndexError:
            print("Error parsing article body: {}".format(self.doi))
            body_word_count = 0
        return body_word_count

    @filename.setter
    def filename(self, value):
        """Sets an article object using a local filename.

        Converts a filename to DOI using an existing function.
        :param value: filename
        :type value: string
        """
        self.doi = filename_to_doi(value)

    @classmethod
    def from_filename(cls, filename):
        """Initiate an article object using a local XML file.

        Will set `self.directory` if the full file path is available. If not, it will
        default to `get_corpus_dir()` via `Article().__init__`. This method is most useful
        for instantiating an Article object when the file is not in the default corpus
        directory, or when changing directories.
        """
        if os.path.isfile(filename):
            directory = os.path.dirname(filename)
        else:
            directory = None
        return cls(filename_to_doi(filename), directory=directory)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Make a SQLite DB out of articles XML files
"""

import argparse
import datetime
import os
from itertools import islice

from tqdm import tqdm
import sqlite3

from peewee import Model, CharField, ForeignKeyField, TextField, \
    DateTimeField, BooleanField, IntegerField, IntegrityError
from playhouse.sqlite_ext import SqliteExtDatabase

from allofplos.corpus import Corpus
from allofplos.transformations import filename_to_doi, convert_country
from allofplos import starterdir
from allofplos.article import Article

journal_title_dict = {
    'PLOS ONE': 'PLOS ONE',
    'PLOS GENETICS': 'PLOS Genetics',
    'PLOS GENET': 'PLOS Genetics',
    'PLOS NEGLECTED TROPICAL DISEASES': 'PLOS Neglected Tropical Diseases',
    'PLOS NEGL TROP DIS': 'PLOS Neglected Tropical Diseases',
    'PLOS BIOLOGY': 'PLOS Biology',
    'PLOS BIOL': 'PLOS Biology',
    'PLOS CLINICAL TRIALS': 'PLOS Clinical Trials',
    'PLOS MEDICINE':'PLOS Medicine',
    'PLOS MEDICIN': 'PLOS Medicine',
    'PLOS MED': 'PLOS Medicine',
    'PLOS PATHOG': 'PLOS Pathogens',
    'PLOS PATHOGENS': 'PLOS Pathogens',
    'PLOS COMPUTATIONAL BIOLOGY': 'PLOS Computational Biology',
    'PLOS COMPUT BIOL': 'PLOS Computational Biology',
    'PLOS CLINICAL TRIALS': 'PLOS Clinical Trials',
}


parser = argparse.ArgumentParser()
parser.add_argument('--db', action='store', help=
                    'Name the db', default='ploscorpus.db')
parser.add_argument('--random', action='store', type=int, help=
                    'Number of articles in a random subset. '
                    'By default will use all articles')
parser.add_argument('--starterdb', action='store_true', help=
                    'Make the starter database', dest='starter')
args = parser.parse_args()

# TODO: Put a warning that the DB will be deleted
if os.path.isfile(args.db):
    os.remove(args.db)

if args.starter:
    if os.path.isfile('starter.db'):
        os.remove('starter.db')
    db = SqliteExtDatabase('starter.db')
else:
    db = SqliteExtDatabase(args.db)

class BaseModel(Model):
    class Meta:
        database = db

class Journal(BaseModel):
    """
    Journal table stores journals values. It is used in the PLOSArticle table.
    Fields:
    journal: Journal name. Based on the values of journal_title_dict
    """
    journal = CharField(unique=True)

class ArticleType(BaseModel):
    """
    ArticleType table stores article types used in PLOSArticle table.
    Fields:
    article_type: Article type such as Research Article, Retraction,
      Essay, Perspective and others.
    """
    article_type = CharField(unique=True)

class Country(BaseModel):
    """
    Country table stores countries that are related with the corresponding
      author.
    Fields:
    country: Country.
    """
    country = CharField(unique=True)

class Affiliations(BaseModel):
    """
    Affiliations table stores affiliations that are related with the
      corresponding author.
    Fields:
    affiliations: Author affiliations such as "Department of Public Health,
      University of Helsinki, Finland".
    """
    affiliations = CharField(unique=True)

class Subjects(BaseModel):
    """
    Subjects table stores subjects that are related to PLOSArticle table by
      using a link table (since an article may have multiple subjects)
    Fields:
    subjects: Subjects like "Molecular Biology", "Cell processes" and so on.
    """
    subjects = CharField(unique=True)

class CorrespondingAuthor(BaseModel):
    """
    CorrespondingAuthor table stores the information of the corresponding
      author.
    Fields:
    corr_author_email: e-mail of the corresponding author
    tld: Top Level Domain, last part of the email, such as .ar, .es, .com.
    given_name: Given name of the author.
    surname: Surname of the author.
    group_name: Name of the group if a group is the author. If not, it is null.
    affiliation: Author affiliation, this is a linked field.
    country: Country of the author, this is a linked field.
    """
    corr_author_email = CharField(unique=True)
    tld = TextField(null=True)
    given_name = TextField(null=True)
    surname = TextField(null=True)
    group_name = TextField(null=True)
    affiliation = ForeignKeyField(Affiliations, related_name='aff')
    country = ForeignKeyField(Country, related_name='aff')

class JATSType(BaseModel):
    """
    JATSType table stores the article type using the JATS (Journal Article
      Tag Suite) standard (https://jats.nlm.nih.gov/).
    Fields:
    jats_type: JATS type, such as "research-article", "discussion", "editorial".
    """
    jats_type = CharField(unique=True)

class PLOSArticle(BaseModel):
    """
    PLOSArticle table is the main table of the database. It stores the articles.
    Fields:
    DOI: DOI for the article. For example: 10.1371/journal.pcbi.0030199
    abstract: Abstract for the article.
    title: Title for the article.
    plostype = Article type (internal PLOS classification), this is a linked
      field.
    journal = Journal name, this is a linked field.
    created_date = Article creation date. For example: 2006-04-11 00:00:00
    word_count = Amount of words in the article.
    JATS_type = Article type (JATS classification), this is a linked field.
    """
    DOI = TextField(unique=True)
    abstract = TextField()
    title = TextField()
    plostype = ForeignKeyField(ArticleType, related_name='arttype')
    journal = ForeignKeyField(Journal, related_name='journals')
    created_date = DateTimeField(default=datetime.datetime.now)
    word_count = IntegerField()
    JATS_type = ForeignKeyField(JATSType, related_name='jats')

class SubjectsPLOSArticle(BaseModel):
    """
    SubjectsPLOSArticle is a link table to relate subject with articles
    Fields:
    subject: This field is linked to Subjects table.
    article: This field is linked to PLOSArticle table.
    """
    subject = ForeignKeyField(Subjects)
    article = ForeignKeyField(PLOSArticle)

class CoAuthorPLOSArticle(BaseModel):
    """
    CoAuthorPLOSArticle is a link table to relate a co-author with a PLOS
      article.
    Fields:
    corr_author: This field is linked to CorrespondingAuthor table.
    article: This field is linked to PLOSArticle table.
    """
    corr_author = ForeignKeyField(CorrespondingAuthor)
    article = ForeignKeyField(PLOSArticle)

db.connect()
db.create_tables([Journal, PLOSArticle, ArticleType, CoAuthorPLOSArticle,
                  CorrespondingAuthor, JATSType, Affiliations, Country,
                  SubjectsPLOSArticle, Subjects])


corpus_dir = starterdir if args.starter else None
all_files = Corpus(corpus_dir)
num_files = len(all_files) if args.random is None else args.random

for article in tqdm(islice(all_files, args.random), total=num_files):
    journal_name = journal_title_dict[article.journal.upper()]
    with db.atomic() as atomic:
        try:
            journal = Journal.create(journal = journal_name)
        except IntegrityError:
            db.rollback()
            journal = Journal.get(Journal.journal == journal_name)
    with db.atomic() as atomic:
        try:
            article_type = ArticleType.create(article_type = article.plostype)
        except IntegrityError:
            db.rollback()
            article_type = ArticleType.get(ArticleType.article_type == article.plostype)
    with db.atomic() as atomic:
        try:
            j_type = JATSType.create(jats_type = article.type_)
        except IntegrityError:
            db.rollback()
            j_type = JATSType.get(JATSType.jats_type == article.type_)
    p_art = PLOSArticle.create(
        DOI=article.doi,
        journal = journal,
        abstract=article.abstract.replace('\n', '').replace('\t', ''),
        title = article.title.replace('\n', '').replace('\t', ''),
        plostype = article_type,
        created_date = article.pubdate,
        word_count=article.word_count,
        JATS_type = j_type)
    # Get subject information
    taxonomy_set = set()
    taxonomy = article.taxonomy
    for values in taxonomy.values():
        for value in values:
            for taxon in value:
                taxonomy_set.add(taxon)
    for taxon in taxonomy_set:
        with db.atomic() as atomic:
            try:
                subject = Subjects.create(subjects = taxon)
            except (sqlite3.IntegrityError, IntegrityError):
                db.rollback()
                subject = Subjects.get(Subjects.subjects == taxon)
        SubjectsPLOSArticle.create(
                subject = subject,
                article = p_art
            )
    if article.authors:
        iterable_authors = article.authors
    else:
        iterable_authors = []
    for auths in iterable_authors:
        if auths['email']:
            with db.atomic() as atomic:
                if auths['affiliations']:
                    author_aff = auths['affiliations'][0]
                else:
                    author_aff = 'N/A'
                try:
                    aff = Affiliations.create(affiliations = author_aff)
                except (sqlite3.IntegrityError, IntegrityError):
                    db.rollback()
                    aff = Affiliations.get(Affiliations.affiliations ==
                                           author_aff)
            with db.atomic() as atomic:
                try:
                    if auths['affiliations'][0] == '':
                        country_from_aff = 'N/A'
                    else:
                        country_from_aff = auths['affiliations'][0].\
                                           split(',')[-1].strip()
                except IndexError:
                    country_from_aff = 'N/A'
                country_from_aff = convert_country(country_from_aff)
                try:
                    country = Country.create(country = country_from_aff)
                except IntegrityError:
                    db.rollback()
                    country = Country.get(Country.country == country_from_aff)

            try:
                co_author = CorrespondingAuthor.create(
                    corr_author_email = auths['email'][0],
                    tld = auths['email'][0].split('.')[-1],
                    given_name = auths['given_names'],
                    surname = auths['surname'],
                    group_name = auths['group_name'],
                    affiliation = aff,
                    country = country
                    )
            except IntegrityError:
                co_author = CorrespondingAuthor.\
                            get(CorrespondingAuthor.corr_author_email == auths['email'][0])
            coauthplosart = CoAuthorPLOSArticle.create(
                    corr_author = co_author,
                    article = p_art
                )



""" Includes all global variables
"""
from collections import OrderedDict
import os

from . import get_corpus_dir

from .plos_regex import validate_filename, validate_doi
from .elements import Journal

# URL bases for PLOS's Solr instances, that index PLOS articles
BASE_URL_API = 'https://api.plos.org/search'

BASE_URL_DOI = 'https://doi.org/'
URL_SUFFIX = '&type=manuscript'
INT_URL_SUFFIX = '.XML'
PREFIX = '10.1371/'
SUFFIX_LOWER = '.xml'
annotation = 'annotation'
correction = 'correction'
ANNOTATION_URL = 'https://journals.plos.org/plosone/article/file?id=10.1371/annotation/'
ANNOTATION_DOI = '10.1371/annotation'
BASE_URL_ARTICLE_LANDING_PAGE = 'https://journals.plos.org/plos{}/article?id={}'
BASE_URL_LANDING_PAGE = 'https://journals.plos.org/{}/'
LANDING_PAGE_SUFFIX = '{}?id={}'
doi_url = 'https://doi.org/'

plos_page_dict = {'article': 'article',
                  'asset': 'article/asset',
                  'articleFigsAndTables': 'article/assets/figsAndTables',
                  'articleAuthors': 'article/authors',
                  'citationDownloadPage': 'article/citation',
                  'downloadBibtexCitation': 'article/citation/bibtex',
                  'downloadRisCitation': 'article/citation/ris',
                  'figuresPage': 'article/figures',
                  'assetFile': 'article/file',
                  'assetXMLFile': 'article/file',
                  'articleMetrics': 'article/metrics',
                  'articleRelated': 'article/related'}


def _get_base_page(journal):
    """Make the base of a PLOS URL journal-specific.

    Defaults to PLOS ONE.

    Use in conjunction with `get_page()` in the Article class.
    """
    journal_map = {'PLOS ONE': 'plosone',
                   'PLOS Computational Biology': 'ploscompbiol',
                   'PLOS Neglected Tropical Diseases': 'plosntds',
                   'PLOS Genetics': 'plosgenetics',
                   'PLOS Pathogens': 'plospathogens',
                   'PLOS Biology': 'plosbiology',
                   'PLOS Medicine': 'plosmedicine',
                   'PLOS Clinical Trials': 'plosclinicaltrials',
                   'PLOS Sustainability and Transformation': 'sustainabilitytransformation',
                   'PLOS Climate': 'climate',
                   'PLOS Water': 'water',
                   'PLOS Global Public Health': 'globalpublichealth',
                   'PLOS Digital Health': 'digitalhealth',
                   'PLOS Mental Health': 'mentalhealth',
                   'PLOS Complex Systems': 'complexsystems',
                   }
    try:
        url = BASE_URL_LANDING_PAGE.format(journal_map[journal])
    except KeyError:
        print('URL error for {}'.format(journal))
        url = BASE_URL_LANDING_PAGE.format('one')

    return url


def filename_to_url(filename):
    """
    Transform filename into a downloadable URL where its XML resides
    Includes transform for the 'annotation' DOIs
    Example:
    filename_to_url('allofplos_xml/journal.pone.1000001.xml') = \
    'https://journals.plos.org/plosone/article/file?id=10.1371/journal.pone.1000001'

    :param filename: string representing a filename
    :return: online location of a PLOS article's XML
    """
    if correction in filename:
        article = 'annotation/' + (filename.split('.', 4)[2])
    else:
        article = os.path.splitext((os.path.basename(filename)))[0]
    doi = PREFIX + article
    return doi_to_url(doi)


def filename_to_doi(filename):
    """
    Transform filename into the article's DOI.
    Includes transform for the 'annotation' DOIs.
    Uses regex to make sure it's a file and not a DOI
    Example:
    filename_to_doi('journal.pone.1000001.xml') = '10.1371/journal.pone.1000001'

    :param filename: relative path to local XML file in the get_corpus_dir() directory
    :return: full unique identifier for a PLOS article
    """
    filename = os.path.basename(filename)
    if not validate_filename(filename):
        raise Exception("Invalid format for PLOS filename: {}".format(filename))
    elif correction in filename:
        article = 'annotation/' + filename.split('.', 4)[2]
        doi = PREFIX + article
    else:
        doi = PREFIX + os.path.splitext(filename)[0]
    return doi


def url_to_path(url, directory=None):
    """
    For a given PLOS URL to an XML file, return the relative path to the local XML file
    Example:
    url_to_path('https://journals.plos.org/plosone/article/file?id=10.1371/journal.pone.1000001') = \
    'allofplos_xml/journal.pone.1000001.xml'
    :param url: online location of a PLOS article's XML
    :param directory: defaults to get_corpus_dir(), containing article files
    :return: relative path to local XML file in the directory
    """
    if directory is None:
        directory = get_corpus_dir()
    annot_prefix = 'plos.correction.'
    if url.startswith(ANNOTATION_URL):
        # NOTE: REDO THIS!
        file_ = os.path.join(directory,
                             annot_prefix +
                             url[url.index(ANNOTATION_DOI + '/')+len(ANNOTATION_DOI + '/'):].
                             replace(URL_SUFFIX, '').
                             replace(INT_URL_SUFFIX, '') + '.xml')
    else:
        file_ = os.path.join(directory,
                             url[url.index(PREFIX)+len(PREFIX):].
                             replace(URL_SUFFIX, '').
                             replace(INT_URL_SUFFIX, '') + '.xml')
    return file_


def url_to_doi(url):
    """
    For a given PLOS URL to an XML file, transform it to the article's DOI
    Example:
    url_to_path('https://journals.plos.org/plosone/article/file?id=10.1371/journal.pone.1000001') = \
    '10.1371/journal.pone.1000001'
    :param url: online location of a PLOS article's XML
    :return: full unique identifier for a PLOS article
    """
    return url[url.index(PREFIX):].rstrip(URL_SUFFIX).rstrip(INT_URL_SUFFIX)


def doi_to_url(doi):
    """
    For a given PLOS DOI, return the PLOS URL to that article's XML file
    Example:
    doi_to_url('10.1371/journal.pone.1000001') = \
    'https://journals.plos.org/plosone/article/file?id=10.1371/journal.pone.1000001'
    :param doi: full unique identifier for a PLOS article
    :return: online location of a PLOS article's XML
    """
    if validate_doi(doi) is False:
        raise Exception("Invalid format for PLOS DOI: {}".format(doi))
    journal = Journal.doi_to_journal(doi)
    base_page = _get_base_page(journal)
    return ''.join([base_page, 'article/file?id=', doi, URL_SUFFIX])


def doi_to_path(doi, directory=None):
    """
    For a given PLOS DOI, return the relative path to that local article
    For DOIs that contain the word 'annotation', searches online version of the article xml to extract
    the journal name, which goes into the filename. Will print DOI if it can't find the journal name
    Uses regex to make sure it's a DOI and not a file
    Example:
    doi_to_path('10.1371/journal.pone.1000001') = 'allofplos_xml/journal.pone.1000001.xml'
    :param doi: full unique identifier for a PLOS article
    :param directory: defaults to get_corpus_dir(), containing article files
    :return: relative path to local XML file
    """
    if directory is None:
        directory = get_corpus_dir()
    if not validate_doi(doi):
        raise Exception("Invalid format for PLOS DOI: {}".format(doi))
    elif doi.startswith(ANNOTATION_DOI):
        article_file = os.path.join(directory, "plos.correction." + doi.split('/')[-1] + SUFFIX_LOWER)
    else:
        article_file = os.path.join(directory, doi.lstrip(PREFIX) + SUFFIX_LOWER)
    return article_file


def convert_country(country):
    """
    For a given country, transform it using one of these rules
    :param country: string with the country name
    :return: string with the normalized country name
    """
    if (country and 'China' in country) or \
            country == 'Chin' or country == 'CHINA':
        country = 'China'
    elif country and 'Brazil' in country or \
            country == 'Brasil' or \
            country == 'ITA - Instituto Tecnologico de Aeronautica (':
        country = 'Brazil'
    elif country and 'Argentina' in country:
        country = 'Argentina'
    elif country == 'Czechia':
        country = 'Czech Republic'
    elif 'Norwegian' in country:
        country = 'Norway'
    elif country and 'United Kingdom' in country:
        country = 'United Kingdom'
    elif country and 'Hong Kong' in country:
        country = 'Hong Kong'
    elif country == 'Cameroun':
        country = 'Cameroon'
    elif (country and 'Chile' in country) or country == 'CHILE':
        country = 'Chile'
    elif (country and 'United States of America' in \
            country) or country == 'United States' or country \
            == 'USA' or 'Florida' in country or \
            'California' in country or\
            country == 'National Reference Centre for' or \
            country == 'United State of America' or \
            country == 'U.S.A.' or \
            country == 'Virginia':
        country = 'United States of America'
    elif country=='Republic of Panamá' or country=='Panamá' or 'Panama' in country:
        country = 'Panama'
    elif 'Canada' in country:
        country = 'Canada'
    elif 'Colombia' in country or country == 'Universidad Aut':
        country = 'Colombia'
    elif 'Spain' in country or country=='España':
        country = 'Spain'
    elif 'Iran' in country:
        country = 'Iran'
    elif 'Saudi Arabia' in country:
        country = 'Saudi Arabia'
    elif 'Italy' in country:
        country = 'Italy'
    elif 'Japan' in country:
        country = 'Japan'
    elif 'Germany' in country:
        country = 'Germany'
    elif 'Luxembourg' in country:
        country = 'Luxembourg'
    elif ('France' in country) or country == 'Marseille':
        country = 'France'
    elif country == 'ROC' or country == 'R. O. C':
        country = 'Taiwan'
    elif country == 'Brasil':
        country = 'Brazil'
    elif country == 'México' or 'Mexico' in country or \
            country == 'Centro de Investigación':
        country = 'Mexico'
    elif 'Slowakia' in country:
        country = 'Slowakia'
    elif country == 'Korea' or 'Republic of Korea' in country:
        country = 'South Korea'
    elif country == 'United Kindgom':
        country = 'United Kingdom'
    elif country and 'Netherlands' in country:
        country = 'Netherlands'
    elif country == 'Commonwealth of Australia' or 'Australia' in country:
        country = 'Australia'
    elif 'Singapore' in country:
        country = 'Singapore'
    elif country and (country[0].isdigit() or country[0] == '+'):
        country = 'N/A'
    return country



"""These functions are for parsing individual elements of the article XML tree

Eventually these functions will probably need to be a class.
"""
import datetime
import difflib
import re
import string

import lxml.etree as et
import unidecode


def parse_article_date(date_element, date_format='%d %m %Y'):
    """
    For an article date element, convert XML to a datetime object.
    :param date_element: An article XML element that contains a date
    :param date_format: string format used to convert to datetime object
    :return: datetime object
    """
    day = ''
    month = ''
    year = ''
    for item in date_element.getchildren():
        if item.tag == 'day':
            day = item.text
        if item.tag == 'month':
            month = item.text
        if item.tag == 'year':
            year = item.text
    if day:
        date = (day, month, year)
        string_date = ' '.join(date)
        date = datetime.datetime.strptime(string_date, date_format)
    elif month:
        # try both numerical & word versions of month
        date = (month, year)
        string_date = ' '.join(date)
        try:
            date = datetime.datetime.strptime(string_date, '%m %Y')
        except ValueError:
            date = datetime.datetime.strptime(string_date, '%B %Y')
    elif year:
        date = year
        date = datetime.datetime.strptime(date, '%Y')
    else:
        print('date error')
        date = ''
    return date


def get_rid_dict(contrib_element):
    """ For an individual contributor, get the list of their associated rids.
    More about rids: https://jats.nlm.nih.gov/archiving/tag-library/1.1/attribute/rid.html
    Used in get_contrib_info().
    :param contrib_element: An article XML element with the tag <contrib>
    :return: dictionary matching each type of rid to its value for that contributor
    """
    rid_dict = {}
    contrib_elements = contrib_element.getchildren()
    # get list of ref-types
    rid_type_list = [el.attrib.get('ref-type', 'fn') for el in contrib_elements if el.tag == 'xref']

    # make dict of ref-types to the actual ref numbers (rids)
    for rid_type in set(rid_type_list):
        rid_list = [el.attrib.get('rid', None) for el in contrib_elements if el.tag == 'xref' and el.attrib.get('ref-type', 'fn') == rid_type]
        rid_dict[rid_type] = rid_list

    return rid_dict


def get_author_type(contrib_element):
    """Get the type of author for a single contributor from their accompanying <contrib> element.
    Authors can be 'corresponding' or 'contributing'. Depending on the paper, some elements have a 
    top-level "corresp" attribute that equal yes; otherwise, corresponding status can be inferred
    from the existence of the <xref> attribute ref-type="corresp"
    :param contrib_element: An article XML element with the tag <contrib>
    :return: author type (corresponding, contributing, None)
    """
    answer_dict = {
        "yes": "corresponding",
        "no": "contributing"
    }

    author_type = None
    if contrib_element.get('contrib-type', None) == 'author':
        corr = contrib_element.get('corresp', None)
        if corr:
            author_type = answer_dict.get(corr, None)
        else:
            temp = get_rid_dict(contrib_element).get('corresp', None)
            if temp:
                author_type = answer_dict.get("yes", None)
            else:
                author_type = answer_dict.get("no", None)

    return author_type


def get_contrib_name(contrib_element):
    """Get the name for a single contributor from their accompanying <contrib> element.
    Also constructs their initials for later matching to article-level dictionaries about
    contributors, including get_aff_dict() and get_fn_dict().
    Can also handle 'collab' aka group authors with a group name but no surname or given names.
    :param contrib_element: An article XML element with the tag <contrib>
    :return: dictionary of a single contributor's given names, surname, initials, and group name
    """
    given_names = ''
    surname = ''

    contrib_name_element = contrib_element.find("name")
    if contrib_name_element is not None:
        for name_element in contrib_name_element.getchildren():
            if name_element.tag == 'surname':
                # for some reason, name_element.text doesn't work for this element
                surname = (et.tostring(name_element,
                                       encoding='unicode',
                                       method='text').rstrip(' ').rstrip('\t').rstrip('\n')
                           or "")
            elif name_element.tag == 'given-names':
                given_names = name_element.text
                if given_names == '':
                    print("given names element.text didn't work")
                    given_names = (et.tostring(name_element,
                                               encoding='unicode',
                                               method='text').rstrip(' ').rstrip('\t').rstrip('\n')
                                   or "")
            else:
                pass
        if given_names or surname:
            # construct initials if either given or surname is present
            try:
                contrib_initials = ''.join([part[0].upper() for part in re.split('[-| |,|\.]+', given_names) if part]) + \
                                  ''.join([part[0] for part in re.split('[-| |,|\.]+', surname) if part[0] in string.ascii_uppercase])
            except IndexError:
                contrib_initials = ''
            contrib_name = dict(contrib_initials=contrib_initials,
                                given_names=given_names,
                                surname=surname)
    else:
        # if no <name> element found, assume it's a collaboration
        contrib_collab_element = contrib_element.find("collab")
        group_name = et.tostring(contrib_collab_element, encoding='unicode')
        group_name = re.sub('<[^>]*>', '', group_name).rstrip('\n')
        if not group_name:
            print("Error constructing contrib_name group element")
            group_name = ''
        contrib_name = dict(group_name=group_name)

    return contrib_name


def get_contrib_ids(contrib_element):
    """Get the ids for a single contributor from their accompanying <contrib> element.
    This will mostly get ORCID IDs, and indicate whetherh they are authenticated.
    For more information of ORCIDs, see https://orcid.org/
    :param contrib_element: An article XML element with the tag <contrib>
    :return: list of dictionaries of ids for that contributor
    """
    id_list = []
    for item in contrib_element.getchildren():
        if item.tag == 'contrib-id':
            contrib_id_type = item.attrib.get('contrib-id-type', None)
            contrib_id = item.text
            contrib_authenticated = item.attrib.get('authenticated', None)
            id_dict = dict(id_type=contrib_id_type,
                           id=contrib_id,
                           authenticated=contrib_authenticated
                           )
            id_list.append(id_dict)

    return id_list


def get_credit_taxonomy(contrib_element):
    """Get the contributor roles from the CREDiT taxonomy element when it is present.
    Right now, this is is equivalent to author roles.
    For more information about this data structure, see http://dictionary.casrai.org/Contributor_Roles
    :param contrib_element: An article XML element with the tag <contrib>
    :return: dictionary of contributor roles for an individual contributor

    """
    credit_dict = {}
    for item in contrib_element.getchildren():
        if item.tag == 'role':
            content_type = item.attrib.get('content-type', None)
            if content_type == 'http://credit.casrai.org/':
                content_type = 'CASRAI CREDiT taxonomy'
            role = item.text
            if not credit_dict.get(content_type, None):
                credit_dict[content_type] = [role]
            else:
                credit_dict[content_type].append(role)
    return credit_dict


def match_contrib_initials_to_dict(contrib_dict, special_dict, matched_keys, contrib_key):
    """For an individual contributor, match their initials to a dictionary.
    This is used for both matching contributors to email addresses as well as credit roles,
    where the keys for all dictionaries are contributor initials. In contrib_dict, these initials are
    constructed from the contributor name in get_contrib_name(). For the special dicts, initials are
    provided in the raw XML.
    See match_contribs_to_dicts() for how this matching process is iterated across contributors.
    :param contrib_dict: information about individual contributor, including their name and constructed initials
    :param special_dict: usually either get_aff_dict() or get_credit_dict()
    :param matched_keys: list of keys in special_dict already matched that will be excluded
    :param contrib_key: The item in the contrib dictionary where the matched special_dict will be stored
    :return: updated contrib_dict that includes the newly matched special_dict
    """
    contributor_initials = contrib_dict.get('contrib_initials')
    # special_dict keys (initials) are assumed to be uppercase
    special_dict = {k.upper(): v
                    for k, v in special_dict.items()
                    if k not in matched_keys}
    if contrib_dict.get('group_name', None) is None:
        try:
            contrib_dict[contrib_key] = special_dict[contributor_initials.upper()]
        except KeyError:
            # Sometimes middle initials are included or excluded, so restrict both initial sets to
            # first and last initial only.
            try:
                contributor_abbrev_initials = ''.join([contributor_initials[0], contributor_initials[-1]])
                for dict_initials, dict_value in special_dict.items():
                    if contributor_abbrev_initials == ''.join([dict_initials[0], dict_initials[-1]]).upper():
                        contrib_dict[contrib_key] = dict_value
                        break
            except (IndexError, KeyError) as e:
                pass

    return contrib_dict


def get_contrib_info(contrib_element):
    """Get a dictionary of information for a single contributor from their accompanying <contrib> element.
    Don't call this function directly. Instead, use as a part of get_contributors_info()
    This includes all contributor information that can be directly accessed from <contrib> element contents.
    However, other contributor information is stored in article-level dictionaries that need to be matched
    for each contributor using the rid_dict created here.
    :param contrib_element: An article XML element with the tag <contrib>
    :return: dictionary of contributor name, ids/ORCID, rid_dict, author_roles
    """
    # get their name
    contrib_dict = get_contrib_name(contrib_element)

    # get contrib type
    try:
        contrib_dict['contrib_type'] = contrib_element.attrib['contrib-type']
    except KeyError:
        # invalid contributor field; shouldn't count as contributor
        return None

    # get author type
    if contrib_dict.get('contrib_type') == 'author':
        contrib_dict['author_type'] = get_author_type(contrib_element)
    elif contrib_dict.get('contrib_type') == 'editor':
        for item in contrib_element.getchildren():
            if item.tag == 'Role' and item.text.lower() != 'editor':
                print('new editor type: {}'.format(item.text))
                contrib_dict['editor_type'] = item.text

    # get ORCID, if available
    contrib_dict['ids'] = get_contrib_ids(contrib_element)

    # get dictionary of contributor's footnote types to footnote ids
    contrib_dict['rid_dict'] = get_rid_dict(contrib_element)

    # get dictionary of CREDiT taxonomy, if available
    contrib_dict['author_roles'] = get_credit_taxonomy(contrib_element)

    return contrib_dict


def match_author_names_to_emails(corr_author_list, email_dict):
    """ Finds the best match of author names to potential matching email addresses.
    Don't call directly, but use as a part of match_contribs_to_dicts().
    Sometimes, the initials-matching process in match_contrib_initials_to_dict() fails. When there's
    at least two ummatched corresponding authors and email addresses, this function
    figures out the name/email matching with the highest matching continguous character count and matches them.
    This is a 'hail Mary' that thankfully also has a high rate of accuracy.
    :param corr_author_list: list of contrib_dicts for corresponding authors with no email field
    :param email_dict: list of unmatched author email addresses
    :return: list of updated contrib_dicts for each author, now including an email field
    """
    overall_matching_dict = {}
    match_values = []
    # Step 1: for each author and email combination, compute longest common string
    for corr_author in corr_author_list:
        # make single string of author full name
        seq_1 = unidecode.unidecode(''.join([corr_author.get('given_names'), corr_author.get('surname')]).lower())
        matching_dict = {}
        for email_initials, email_address in email_dict.items():
            # make string of email address that doesn't include domain
            seq_2 = unidecode.unidecode(email_address[0].lower().split('@')[0])
            matcher = difflib.SequenceMatcher(a=seq_1, b=seq_2)

            # construct dictionary with name, email, and matching string length for each pair
            match = matcher.find_longest_match(0, len(matcher.a), 0, len(matcher.b))
            matching_dict[tuple(email_address)] = match[-1]
            # add length of match to list of all match lengths
            match_values.append(match[-1])
        overall_matching_dict[(corr_author.get('given_names'), corr_author.get('surname'))] = matching_dict
    # Step 2: for the author and email combination(s) with the longest common string, match them
    # Iterate through max_values in descending order until all are matched
    newly_matched_emails = []
    newly_matched_authors = []
    count = 0
    while len(newly_matched_emails) < len(overall_matching_dict) and count < 20:
        for k1, v1 in overall_matching_dict.items():
            for k2, v2 in v1.items():
                if v2 == max(match_values):
                    for corr_author in corr_author_list:
                        if k1 == (corr_author.get('given_names'), corr_author.get('surname')) \
                         and k2 not in newly_matched_emails and k1 not in newly_matched_authors:
                            corr_author['email'] = list(k2)
                            # keep track of matched email & author so they're not matched again
                            newly_matched_authors.append(k1)
                            newly_matched_emails.append(k2)
                            match_values.remove(v2)
            count += 1
    # Step 3: match the remaining author and email if there's only one remaining (most common)
    # Might not be necessary with the while loop
    still_unmatched_authors = [author for author in corr_author_list if not author.get('email')]
    still_unmatched_emails = {k: v for k, v in email_dict.items() if tuple(v) not in newly_matched_emails}
    if len(still_unmatched_authors) == len(still_unmatched_emails) <= 1:
        if len(still_unmatched_authors) == len(still_unmatched_emails) == 1:
            # only one remaining. it gets matched
            still_unmatched_authors[0]['email'] = list(still_unmatched_emails.values())[0]
        else:
            # we were successful at matching all emails (likely, two pairs had the same match values)
            pass
    else:
        # something's gone wrong. the candidate list of emails doesn't match the number of authors
        # the corresponding authors printed below will have their ['email'] field unfilled
        print('not calculating right', still_unmatched_authors, still_unmatched_emails)

    return corr_author_list


def match_contribs_to_dicts(contrib_list, special_dict, contrib_key):
    """
    :param contrib_list: list of contributors
    :param special_dict: usually either get_aff_dict() or get_credit_dict()
    :param contrib_key: The item in the contrib dictionary where the matched special_dict will be stored
    """
    matching_error = False
    matched_keys = []
    for contrib_dict in contrib_list:
        contrib_dict = match_contrib_initials_to_dict(contrib_dict,
                                                      special_dict,
                                                      matched_keys,
                                                      contrib_key)
        if contrib_dict.get(contrib_key, None):
            for k, v in special_dict.items():
                if v == contrib_dict.get(contrib_key):
                    matched_keys.append(k)
    if len(special_dict) == len(matched_keys):
        # all special_dicts and contributors are matched
        pass
    else:
        unmatched_special_dict = {k: v for k, v in special_dict.items()
                                  if k not in matched_keys}
        contrib_dict_missing_special_list = [contrib_dict for contrib_dict in contrib_list
                                             if not contrib_dict.get(contrib_key, None)]

        # if one contributor and one special_dict are unmatched, match them
        if len(unmatched_special_dict) == len(contrib_dict_missing_special_list) == 1:
            contrib_dict_missing_special_list[0][contrib_key] = list(unmatched_special_dict.values())[0]

        elif len(unmatched_special_dict) != len(contrib_dict_missing_special_list):
            # these numbers should always be the same
            matching_error = True

        else:
            if contrib_key == 'email':
                # match remaining contributor names to emails by string matching
                contrib_dicts = match_author_names_to_emails(contrib_dict_missing_special_list, unmatched_special_dict)
                if len([contrib for contrib in contrib_dicts if contrib_key not in contrib.keys()]) == 0:
                    # finally every contributor and special_dict is matched
                    pass
                else:
                    # even after applying every strategy, there were unmatched contributors
                    matching_error = True
    return contrib_list, matching_error



import textwrap

def dedent(text):
    """Equivalent of textwrap.dedent that ignores unindented first line.
    This means it will still dedent strings like:
    '''foo
    is a bar
    '''
    For use in wrap_paragraphs.
    
    Taken from https://github.com/ipython/ipython_genutils/text.py
    """

    if text.startswith('\n'):
        # text starts with blank line, don't ignore the first line
        return textwrap.dedent(text)

    # split first line
    splits = text.split('\n',1)
    if len(splits) == 1:
        # only one line
        return textwrap.dedent(text)

    first, rest = splits
    # dedent everything but the first line
    rest = textwrap.dedent(rest)
    return '\n'.join([first, rest])



"""
The following RegEx pertains to the 7 main PLOS journals and the defunct PLOS Clinical Trials, as well as PLOS Currents.
"""

import re
import os

from . import get_corpus_dir

regex_match_prefix = r"^10\.1371/"
regex_body_match = (r"((journal\.p[a-zA-Z]{3}\.[\d]{7}$)"
                    r"|(annotation/[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}$))")
regex_body_search = (r"((journal\.p[a-zA-Z]{3}\.[\d]{7})"
                     r"|(annotation/[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}))")
regex_body_currents = (r"((currents\.[a-zA-Z]{2,9}\.[a-zA-Z0-9]{32}$)"
                       r"|(currents\.RRN[\d]{4}$)"
                       r"|([a-zA-Z0-9]{13}$)"
                       r"|([a-zA-Z0-9]{32}$))")
regex_file_search = (r"((journal\.p[a-zA-Z]{3}\.[\d]{7})"
                     r"|(plos\.correction\.[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}))")
full_doi_regex_match = re.compile(regex_match_prefix+regex_body_match)
full_doi_regex_search = re.compile(r"10\.1371/journal\.p[a-zA-Z]{3}\.[\d]{7}"
                                   "|10\.1371/annotation/[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}")
currents_doi_regex = re.compile(regex_match_prefix+regex_body_currents)
file_regex_match = re.compile(regex_file_search+r"\.xml")
BASE_URL = 'https://journals.plos.org/plosone/article/file?id='
URL_SUFFIX = '&type=manuscript'
external_url_regex_match = re.compile(re.escape(BASE_URL) +
                                      re.escape("10.1371/") +
                                      regex_body_search +
                                      re.escape(URL_SUFFIX))


def validate_doi(doi):
    """
    For an individual string, tests whether the full string is in a valid PLOS DOI format or not
    Example: '10.1371/journal.pbio.2000777' is True, but '10.1371/journal.pbio.2000777 ' is False
    :return: True if string is in valid PLOS DOI format; False if not
    """
    return bool(full_doi_regex_match.search(doi))


def validate_filename(filename):
    """
    For an individual string, tests whether the full string is in a valid article file. This can take two forms.
    
    TODO: Officially document these two forms and give them names. Also, Explain the example below.
    
    Example: 'allofplos_xml/journal.pbio.2000777.xml' is True, but 'allofplos_xml/journal.pbio.20007779.xml' is False
    :filename: A string with a file name
    :return: True if string is in a valid PLOS corpus article format; False if not
    """
    if file_regex_match.search(filename):
        return True
    else:
        return False


def validate_url(url):
    """
    For an individual string, tests whether the full string is in a valid article url format or not
    Example: 'https://journals.plos.org/plosone/article/file?id=10.1371/journal.pcbi.0020147&type=manuscript' is True,
    but 'https://journals.plos.org/plosone/article/file?id=10.1371/journal.pcbi.0020147' is False
    :return: True if string is in a valid PLOS article url; False if not
    """
    return bool(external_url_regex_match.search(url))


def find_valid_dois(doi):
    """
    For an individual string, searches for any valid PLOS DOIs within it and returns them
    :return: list of valid PLOS DOIs contained within string
    """
    return full_doi_regex_search.findall(doi)


def show_invalid_dois(doi_list):
    """
    Checks to see whether a list of PLOS DOIs follow the correct format. Used mainly to determine
    if linked DOI fields in other articles (such as retractions and corrections) are correct.
    :return: list of DOI candidates that don't match PLOS's pattern
    """
    return list(filter(lambda x: not validate_doi(x), doi_list))


def currents_doi_filter(doi_list):
    """
    Checks to see whether a list of PLOS Currents DOIs follow the correct format. Used mainly to determine
    if linked DOI fields in PMC articles are correct.
    :return: list of DOI candidates that don't match Currents' pattern
    """
    return list(filter(lambda x: not bool(currents_doi_regex.search(x)), doi_list))



import tempfile
import os

# path to the root of allofplos (the package)
ALLOFPLOS_DIR_PATH = os.path.abspath(os.path.dirname(__file__))


# Starter pack of PLOS articles
starterdir = os.path.join(ALLOFPLOS_DIR_PATH, 'starter_corpus')

# Temporary folder for downloading and processing new articles
newarticledir = tempfile.mkdtemp()

# List of uncorrected proof articles to check for updates
(_, uncorrected_proofs_text_list) = tempfile.mkstemp()

def get_corpus_dir():
    """If you want to set the corpus directory, assign the desired path to
    ``os.environ['PLOS_CORPUS']``.
    """
    import os

    return os.path.expanduser(os.environ.get("PLOS_CORPUS", "")) or os.path.join(
        ALLOFPLOS_DIR_PATH, "allofplos_xml"
    )

del os

    

# NB: any packages that you want to expose at the top level, you will need to
# import after creating global variables that they may rely upon
# (e.g., corpusdir)

from .article import Article
from .corpus import Corpus



""" This is an example on how to convert the SQLite DB into a class model to
    make queries without using SQL syntax. This example uses peewee as ORM.
    It should be used as a template, the end user could edit it to fit their
    needs.
"""

from peewee import *

# Start of ORM classes creation.

database = SqliteDatabase('starter.db', **{})

class UnknownField(object):
    def __init__(self, *_, **__): pass

class BaseModel(Model):
    class Meta:
        database = database

class Affiliations(BaseModel):
    affiliations = CharField(unique=True)

    class Meta:
        db_table = 'affiliations'

class Articletype(BaseModel):
    article_type = CharField(unique=True)

    class Meta:
        db_table = 'articletype'

class Jatstype(BaseModel):
    jats_type = CharField(unique=True)

    class Meta:
        db_table = 'jatstype'

class Journal(BaseModel):
    journal = CharField(unique=True)

    class Meta:
        db_table = 'journal'

class Plosarticle(BaseModel):
    doi = TextField(db_column='DOI', unique=True)
    jats_type = ForeignKeyField(column_name='JATS_type_id', model=Jatstype, field='id')
    abstract = TextField()
    created_date = DateTimeField()
    journal = ForeignKeyField(column_name='journal_id', model=Journal, field='id')
    plostype = ForeignKeyField(column_name='plostype_id', model=Articletype, field='id')
    title = TextField()
    word_count = IntegerField()

    class Meta:
        db_table = 'plosarticle'

class Country(BaseModel):
    country = CharField(unique=True)

    class Meta:
        db_table = 'country'

class Correspondingauthor(BaseModel):
    affiliation = ForeignKeyField(column_name='affiliation_id', model=Affiliations, field='id')
    corr_author_email = CharField(unique=True)
    country = ForeignKeyField(column_name='country_id', model=Country, field='id')
    given_name = TextField(null=True)
    group_name = TextField(null=True)
    surname = TextField(null=True)
    tld = TextField(null=True)

    class Meta:
        db_table = 'correspondingauthor'

class Coauthorplosarticle(BaseModel):
    article = ForeignKeyField(column_name='article_id', model=Plosarticle, field='id')
    corr_author = ForeignKeyField(column_name='corr_author_id', model=Correspondingauthor, field='id')

    class Meta:
        db_table = 'coauthorplosarticle'

# End of ORM classes creation.

# Query the starter database to retrieve all paper published in the journal
# PLOS Computational Biology, since 2008 with a corresponding author from
# France.
query = (Plosarticle
         .select()
         .join(Coauthorplosarticle)
         .join(Correspondingauthor)
         .join(Country)
         .join(Journal, on=(Plosarticle.journal == Journal.id))
         .where(Country.country == 'France')
         .where(Plosarticle.created_date > '2008-1-1')
         .where(Journal.journal == 'PLOS Computational Biology')
         )

# Get how many papers are returned
print("Papers: {}".format(query.count()))

# Get the DOIs of all the papers found with the query
print("DOIs:")
for papers in query:
    print(papers.doi)



import os

from random import Random
from collections import OrderedDict
from itertools import islice

from .. import get_corpus_dir, Article
from ..transformations import filename_to_doi, doi_to_path


class Corpus:
    """A collection of PLOS articles."""

    def __init__(self, directory=None, extension='.xml', seed=None):
        """Creation of an article corpus class."""
        if directory is None:
            directory = get_corpus_dir()
        self.directory = directory
        self.extension = extension
        self.random = Random(seed)

    def __repr__(self):
        """Value of a corpus object when you call it directly on the command line.

        Shows the directory location of the corpus
        :returns: directory
        :rtype: {str}
        """
        out = "Corpus location: {0}\nNumber of files: {1}".format(self.directory, len(self.files))
        return out
    
    def __len__(self):
        return len(self.dois)
    
    def __iter__(self):
        return (article for article in self.random_article_generator)
    
    def __getitem__(self, key):
        
        if isinstance(key, int):
            return Article(self.dois[key], directory=self.directory)
        elif isinstance(key, slice):
            return (Article(doi, directory=self.directory) 
                    for doi in self.dois[key])
        elif key not in self.dois:
            path= doi_to_path(key, directory=self.directory)
            raise IndexError(("You attempted get {doi} from "
                              "the corpus at \n{directory}. \n"
                              "This would point to: {path}. \n"
                              "Is that the file that was intended?"
                              ).format(doi=key, 
                                       directory=self.directory,
                                       path=path
                                      )
                            )
        else:
            return Article(key, directory=self.directory)

    def __contains__(self, value):
        is_in = False
        if isinstance(value, Article):
            is_in = value.doi in self.dois and value.directory == self.directory
        elif isinstance(value, str):
            doi_in = value in self.dois
            file_in = value in self.files
            filepath_in = value in self.filepaths
            is_in = doi_in or file_in or filepath_in
        return is_in
        
    @property
    def iter_file_doi(self):
        """Generator that returns filename, doi tuples for every file in the corpus.

        Used to generate both DOI and file generators for the corpus.
        """
        return ((file_, filename_to_doi(file_))
                for file_ in sorted(os.listdir(self.directory))
                if file_.endswith(self.extension) and 'DS_Store' not in file_)

    @property
    def file_doi(self):
        """An ordered dict that maps every corpus file to its accompanying DOI."""
        return OrderedDict(self.iter_file_doi)

    @property
    def iter_files(self):
        """Generator of article XML filenames in the corpus directory."""

        return (x[0] for x in self.iter_file_doi)

    @property
    def iter_dois(self):
        """Generator of DOIs of the articles in the corpus directory.

        Use for looping through all corpus articles with the Article class.
        """

        return (x[1] for x in self.iter_file_doi)

    @property
    def iter_filepaths(self):
        """Generator of article XML files in corpus directory, including the full path."""
        return (os.path.join(self.directory, fname) for fname in self.iter_files)

    @property
    def files(self):
        """List of article XML files in the corpus directory."""

        return list(self.iter_files)

    @property
    def dois(self):
        """List of DOIs of the articles in the corpus directory."""

        return list(self.iter_dois)

    @property
    def filepaths(self):
        """List of article XML files in corpus directory, including the full path."""
        return list(self.iter_filepaths)

    @property
    def article_generator(self):
        """iterator of articles"""
        return (Article(doi, directory=self.directory) 
                for doi in self.iter_dois)

    @property
    def random_article_generator(self):
        """iterator over random articles"""
        return (Article(doi, directory=self.directory) 
                for doi in self.iter_random_dois)

    @property
    def random_article(self):
        return next(self.random_article_generator)

    def random_sample(self, count):
        """
        Creates a generator for random articles. 

        Length of generator specified in `count` parameter.

        :param count: specify how many articles are to be returned
        :return: a generator of random articles for analysis
        """

        return islice(self.random_article_generator, count)

    @property
    def iter_random_dois(self):
        return (doi for doi in self.random.sample(self.dois, len(self)))

    @property
    def random_doi(self):
        return next(self.iter_random_dois)

    def random_dois(self, count):
        """
        Gets a list of random DOIs found in the corpus.

        Length of list specified in `count` parameter.
        :param count: specify how many DOIs are to be returned
        :return: a list of random DOIs for analysis
        """

        return list(islice(self.iter_random_dois, count))



# Copyright (c) 2014 Public Library of Science

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# !/usr/bin/env python
#
# Code written by Adam Becker
# Code refactored into Python 3 by Elizabeth Seiver


"""
citation_utilities.py
Version 0.10, 2014-01-24
"""


from numpy import median
from bs4 import BeautifulSoup
from itertools import chain, compress
import requests
from urllib.parse import quote
import json
import re
from multiprocessing import Pool

from tqdm import tqdm


def soupify(filename):
    '''Opens the given XML file, parses it using Beautiful Soup, and returns the output.'''
    f = open(filename, "r")
    soup = BeautifulSoup(f, features="xml")
    f.close()
    return soup


def remote_retrieve(doi, filename=''):
    '''Given the DOI of a PLOS paper, downloads the XML.'''
    # First, see if the file was already downloaded.
    if filename:
        try:
            open(filename)
            return filename
        except IOError:
            pass
    headers = {"Content-Type": "application/xml"}
    r = requests.get("https://www.plosone.org/article/fetchObjectAttachment.action?uri=info:doi/" + doi + "&representation=XML", headers=headers)
    # Doesn't matter whether it's a PLOS ONE article or not -- this will work for any article in any PLOS journal.
    r.encoding = "UTF-8"  # This is needed to keep the encoding on the papers correct.
    if filename:
        f = open(filename, "w")
        f.write(r.text.encode("UTF-8"))
        f.close()
        return filename
    # Nothing after a return statement is executed, so the following line will only happen if filename is False.
    return r.text


def remote_soupify(doi):
    '''Given the DOI of a PLOS paper, downloads the XML and parses it using Beautiful Soup.'''
    headers = {"Content-Type": "application/xml"}
    r = requests.get("https://www.plosone.org/article/fetchObjectAttachment.action?uri=info:doi/" + doi + "&representation=XML", headers=headers)
    # Doesn't matter whether it's a PLOS ONE article or not -- this will work for any article in any PLOS journal.
    r.encoding = "UTF-8"  # This is needed to keep the encoding on the papers correct.
    soup = BeautifulSoup(r.text, features="xml")
    return soup


def citation_grouper(paper):
    '''
    Don't call this directly -- use one of the dictionary or database functions instead.
    Given a soupified XML paper, returns a list of the 'citation groups' within that paper.
    A citation group is all of the papers that are cited together at a certain point in the paper -- the bare stuff of inline co-citations.
    So, for example, a citation group might simply be [1], or it could be [2]-[10], or even [11], [13]-[17], [21].
    The elements within the returned list are themselves lists of soupified XML elements corresponding to citations and connectives.
    '''

    # Make a place for the output to live.
    groups = []

    # Find the first citation.
    cite = paper.find(attrs={"ref-type":"bibr"})

    # This is the loop that actually goes through and finds the inline co-citation groups.
    # It'll continue as long as the "cite" variable isn't None, False, or 0.
    while cite:
        # Make a place for the citation group to live, and add the first member of the group.
        group = [cite]

        # Get the next sibling on the XML tree,
        # determine whether it's part of the group or not,
        # rinse & repeat.
        next = cite.next_sibling
        while next and ((next.name == "xref" and next["ref-type"] == "bibr") or ((next == ", " or next == "," or next == "\u2013" or next == "-" or next == "\u2014" or next == " ") and (next.next_sibling and next.next_sibling.name == "xref" and next.next_sibling["ref-type"] == "bibr"))):
            group.append(next)
            next = next.next_sibling

        # Put the group into the list of groups, then find the next citation.
        groups.append(group)
        try:
            cite = next.find_next(attrs={"ref-type": "bibr"})  # At the end of the document, this will return None, breaking the loop condition.
        except AttributeError:
            cite = cite.find_next(attrs={"ref-type": "bibr"})
            # we need this try-except clause in the event that the citation has no further siblings, meaning next will be None and will fail to have the find_next method.

    return groups


def number(citation):
    '''A little function that translates XML citations into their citation numbers.'''
    # I attempted to make this function more subtle and capable of handling a wider variety of citation styles.
    # I failed! But the attempt is instructive, so it's commented out below.
    #
    # try:
    #     # Sometimes, the citation tag is inside the brackets, eg. [<xref>12</xref>]
    #     naive = int(citation.text)
    #     return naive
    # except ValueError:
    #     try:
    #         # But sometimes, the brackets are inside the tag.
    #         less_naive = int(citation.text[1:-1])
    #         return less_naive
    #     except ValueError:
    #         # And sometimes, there's a full-text citation rather than a number! (Only on older papers, I think.)
    #         rid = citation["rid"]
    #         ref = citation.find_next("ref", attrs = {"id": rid})
    #         label = ref.label
    #         # number = re.search(r"\d+", citation.text)
    #         number = re.search(r"\d+", label.text)
    #         return int(number.group())
    #
    # The real function is below.
    # This just pulls the reference ID off of the citation's XML tags,
    # then looks up the reference number of that RID down in the references section of the paper.
    rid = citation["rid"]
    ref = citation.find_next("ref", attrs={"id": rid})
    label = ref.label
    try:
        number = re.search(r"\d+", label.text)
        return int(number.group())
    except AttributeError:
        print("Missing reference labels on DOI " + ref.find_previous("article-id", attrs={"pub-id-type": "doi"}).text.encode("UTF-8") + " for reference " + citation.text.encode("UTF-8") + ".")
        return 0  # This should be a reasonable thing to return in the case that no citation number is found: it won't cause errors elsewhere, but it's easy to track.


def group_cleaner(group):
    '''
    Don't call this directly -- use one of the dictionary or database functions instead.
    Given a citation group like the ones in the list returned by citation_grouper,
    returns lists of integers that correspond to citation numbers.
    So XML corresponding to [3], [5]-[7] would be returned as [3, 5, 6, 7].
    '''
    # The idea here relies (as it almost has to)
    # on the fact that the citation groups have citations alternating with punctuation
    # (that, or they're just a single citation).
    # So we start with a citation, then punctuation, then citation, and so on,
    # always starting and ending with a citation -- and sometimes that's the same one, the only member of the group.

    # Make a place for the output to live.
    cite_numbers = []

    # Get the first element
    first_citation = group[0]
    old_number = number(first_citation)
    cite_numbers.append(old_number)
    # Set up the loop.
    hyphen_bool = False

    for thing in group[1:]:
        if thing.name == "xref":
            new_number = number(thing)
            if hyphen_bool:
                cite_numbers.extend(list(range(old_number + 1, new_number)))
            cite_numbers.append(new_number)
            old_number = new_number
        else:
            # Figure out whether the punctuation is a hyphen or a comma, and act accordingly.
            if thing == "\u2013" or thing == "-" or thing == "\u2014":
                hyphen_bool = True
            else:
                hyphen_bool = False

    return cite_numbers


def doi(number, paper, verbose=True):
    '''
    DEPRECATED in favor of doi_batch.
    Given a soupified paper and a citation number, attempts to return the DOI of the reference
    by searching the text of the reference and doing a CrossRef search if that fails.
    Returns None if no DOI can be found.
    Verbose does what it sounds like.
    NB: The automated CrossRef search has intermittent trouble finding things.
    Not sure why, but this function definitely fails to find DOIs that are nonetheless in the CrossRef database.
    '''
    # Get the references.
    references = paper.find("ref-list")
    ref_label = references.find("label", text=str(number))
    ref = ref_label.find_next_sibling("mixed-citation").text
    # Try searching for an inline DOI first.
    doimatch = re.search(r"\sdoi:|\sDOI:|\sDoi:|\.doi\.|\.DOI\.", ref)
    if doimatch:
        rawdoi = ref[doimatch.start():]
        doi = rawdoi[rawdoi.index("10."):]  # all DOI's start with 10., see reference here: http://www.doi.org/doi_handbook/2_Numbering.html#2.2

        # Removing whitespace and anything afterwards.
        space = re.search(r"\s", doi)
        if space:
            doi = doi[:space.start()]

        # Removing trailing periods.
        if doi[-1] == ".":
            doi = doi[:-1]

        return doi
    else:
        # Now search for the DOI on Crossref.
        if verbose:
            print("No inline DOI found for this reference. Trying CrossRef search...")
        url = "http://search.crossref.org/links"
        data = json.dumps([ref])
        headers = {"Content-Type": "application/json"}
        r = requests.post(url, data=data, headers=headers)
        if r.json()["query_ok"]:
            results = r.json()["results"][0]
            if results["match"]:
                rawdoi = results["doi"]
                doi = rawdoi[rawdoi.index("10"):]  # CrossRef returns DOIs of the form http://dx.doi/org/10.<whatever>
                return doi
            else:
                if verbose:
                    print("Couldn't find a DOI, sorry!")
        else:
            print("There's a problem with the CrossRef DOI search. Check your internet connection and confirm the original paper was properly formatted in the PLOS XML style, then try again.")
        return None


def doi_batch(paper, crossref=False):
    '''
    Returns all the dois for a whole paper, in one batch.
    Works somewhat like the individual doi function above -- searches for a doi inline, then looks elsewhere if that fails --
    but this function looks at the inline HTML DOIs on the PLOS website for the DOIs by default.
    If crossref=True, it uses CrossRef instead, but it submits all the CrossRef requests at once,
    so it isn't spamming the CrossRef server the way a long series of individual doi calls would.
    '''
    # Get the doi of the given paper.
    paper_doi = plos_paper_doi(paper)
    # Find all the references.
    references = paper.find_all("ref")
    max_ref_num = len(references)
    ref_nums = list(range(1, max_ref_num + 1))
    refs = {i: r.text for i, r in zip(ref_nums, references)}
    dois = {}
    cr_queries = {}
    # Try searching for inline DOIs first.
    for i, ref in refs.items():
        doimatch = re.search(r"doi:|DOI:|Doi:|doi|DOI", ref)
        if doimatch:
            rawdoi = ref[doimatch.start():]
            try:
                doi = rawdoi[rawdoi.index("10."):]
                # all DOI's start with 10., see reference here: http://www.doi.org/doi_handbook/2_Numbering.html#2.2
            except ValueError:
                # if a ValueError is raised, that means the DOI doesn't contain the string '10.' -- which means it's not a valid DOI.
                cr_queries[i] = ref
                continue  # jump to the next reference

            # Removing whitespace and anything afterwards.
            space = re.search(r"\s", doi)
            if space:
                doi = doi[:space.start()]

            # Removing trailing periods.
            if doi[-1] == ".":
                doi = doi[:-1]

            dois[i] = doi
        else:
            cr_queries[i] = ref

    print(dois)

    if crossref:
        # Now search for the DOIs on Crossref.
        url = "http://search.crossref.org/links"
        data = json.dumps(list(cr_queries.values()))
        headers = {"Content-Type": "application/json"}
        r = requests.post(url, data=data, headers=headers)
        if r.json()["query_ok"]:
            results = r.json()["results"]
        else:
            print("There's a problem with the CrossRef DOI search. Check your internet connection and confirm the original paper was properly formatted in the PLOS XML style, then try again.")
            return None

        for i, result in zip(list(cr_queries.keys()), results):
            if result["match"]:
                rawdoi = result["doi"]
                doi = rawdoi[rawdoi.index("10"):]  # CrossRef returns DOIs of the form http://dx.doi/org/10.<whatever>
                dois[i] = doi
            else:
                dois[i] = None
    else:
        paper_url = "https://www.plosone.org/article/info:doi/" + paper_doi
        paper_request = requests.get(paper_url)
        paper_html = BeautifulSoup(paper_request.content, "lxml")
        html_references = paper_html.select('.references > li')

        for i in cr_queries.keys():
            ref = str(html_references[i-1])
            try:
                doimatch = re.search(r"doi:|DOI:|Doi:|doi|DOI", ref)
            except TypeError:
                # if a ValueError is raised, that means that somehow, ref is not a string, which means there's something really weird with the reference.
                # Skip it and go to the next one.
                dois[i] = None
                continue  # jump to the next reference
            if doimatch:
                rawdoi = ref[doimatch.start():]
                try:
                    doi = rawdoi[rawdoi.index("10."):]
                    # all DOI's start with 10., see reference here: http://www.doi.org/doi_handbook/2_Numbering.html#2.2
                except ValueError:
                    # if a ValueError is raised, that means the DOI doesn't contain the string '10.' -- which means it's not a valid DOI.
                    dois[i] = None
                    continue  # jump to the next reference

                # Removing whitespace and anything afterwards.
                space = re.search(r"\s", doi)
                if space:
                    doi = doi[:space.start()]

                # Removing trailing periods.
                if doi[-1] == ".":
                    doi = doi[:-1]

                dois[i] = doi
            else:
                dois[i] = None
    return dois


def intra_paper_mentions(number, paper):
    '''
    DEPRECATED in favor of ipm_dictionary
    Given a soupified paper and a citation number, returns the number of times that citation number is mentioned in the given paper.
    '''
    groups = [group_cleaner(g) for g in citation_grouper(paper)]
    flat_list = list(chain(*groups))
    return flat_list.count(number)
    # And that's all, folks!


def ipm_dictionary(paper):
    '''
    Creates a dictionary of the number of intra-paper mentions for each thing cited in the given paper.
    Basically runs intra_paper_mentions on every cited thing within the given paper, then stuffs them in a database.
    MUCH faster than actually doing that, though.
    '''
    groups = [group_cleaner(g) for g in citation_grouper(paper)]
    flat_list = list(chain(*groups))
    # Find all the references.
    references = paper.find_all("ref")
    max_ref_num = len(references)
    results = {i: flat_list.count(i) for i in range(1, max_ref_num + 1)}
    return results


def ipm_histogram(paper, details=False):
    '''
    Returns a database that is essentially the inverse of what ipm_dictonary spits out:
    now the keys are the number of mentions in the paper, and the values are the number of cited things that are mentioned that many times.
    So d[1] is the number of cited things only mentioned once, etc.
    If details = True, the dictionary will instead return a list of the things with the given number of mentions.
    '''
    cite_dict = ipm_dictionary(paper)
    frequencies = set(cite_dict.values())

    if not details:
        hist = {f: 0 for f in frequencies}  # dict comprehensions FTW!
        for k in list(cite_dict.keys()):
            v = cite_dict[k]
            hist[v] += 1
    else:
        hist = {f: [] for f in frequencies}
        for k in list(cite_dict.keys()):
            v = cite_dict[k]
            hist[v].append(k)

    return hist


def micc(number, paper):
    '''
    DEPRECATED in favor of micc_dictionary
    Given a paper and a citation number within that paper,
    returns the median number of inline co-citations alongside that citation in that paper.
    '''
    # Fairly straightforward.
    all_groups = [group_cleaner(g) for g in citation_grouper(paper)]
    counts = [g.count(number) for g in all_groups]
    cite_groups = compress(all_groups, counts)
    cocite_counts = [len(g) - 1 for g in cite_groups]
    return median(cocite_counts)


def micc_dictionary(paper):
    '''
    Analogous to citation_number_dictionary, but for MICCs rather than the number of citations.
    Co-citations are when two citations are included in the same end note (e.g, '[3-5]')
    :return: dict of counts for co-citation occurrences
    '''
    all_groups = [group_cleaner(g) for g in citation_grouper(paper)]
    references = paper.find_all("ref")
    max_ref_num = len(references)
    results = {}
    for i in range(1, max_ref_num + 1):
        counts = [g.count(i) for g in all_groups]
        cite_groups = compress(all_groups, counts)
        cocite_counts = [len(g) - 1 for g in cite_groups]
        if len(cocite_counts) == 0:
            cocite_counts = [-1]
        results[i] = median(cocite_counts)

    return results


def micc_histogram(paper, details=False):
    '''
    Analogous to citation_histogram, but for MICCs rather than the number of citations.
    '''
    micc_dict = micc_dictionary(paper)
    miccs = set(micc_dict.values())

    if not details:
        hist = {m: 0 for m in miccs}
        for k in list(micc_dict.keys()):
            v = micc_dict[k]
            hist[v] += 1
    else:
        hist = {m: [] for m in miccs}
        for k in list(micc_dict.keys()):
            v = micc_dict[k]
            hist[v].append(k)

    return hist


def citation_database(papers, verbose=True):
    '''
    Given a list of soupified papers in PLOS XML format, assembles a database of the papers those papers cite,
    collects citation numbers and MICCs for each of those papers,
    and then calculates the median number of intra-paper mentions and median MICCs for each paper in the database.
    Returns a dictionary of these measures, along with bare number of citations, keyed by DOI.
    Papers without discoverable DOIs are removed from the database.
    '''
    if verbose:
        print("Processing " + str(len(papers)) + " papers...")

    database = {}

    for i, paper in enumerate(tqdm(papers)):
        if verbose:
            paper_doi = plos_paper_doi(paper)
            print("DOI of paper " + str(i + 1) + " is " + paper_doi)
        # Get the DOIs, the intra-paper mention counts, and the miccs.
        ipms = ipm_dictionary(paper)  # -> citation_counts
        miccs = micc_dictionary(paper)
        if verbose:
            print("Retrieving DOIs for paper " + str(i + 1) + "...")
        dois = doi_batch(paper)

        # Remove the papers with un-identifiable DOIs.
        dois = {k: v for k, v in compress(list(dois.items()), list(dois.values()))}
        if verbose:
            print("Retrieved " + str(len(dois)) + " DOIs.")
            print("Processing database entries...")
        for j, doi in dois.items():
            try:
                database[doi]["ipms"].append(ipms[j])
                database[doi]["miccs"].append(miccs[j])
                database[doi]["citations"] += 1
            except KeyError:
                database[doi] = {}
                database[doi]["ipms"] = [ipms[j]]
                database[doi]["miccs"] = [miccs[j]]
                database[doi]["citations"] = 1

    if verbose:
        print("Database post-processing...")
    for info in database.values():
        info["median_ipm"] = median(info["ipms"])
        info["median_micc"] = median(info["miccs"])

    return database


def paper_citations(filename, verbose=False):
    '''
    Somewhat similar to citation_database.
    Useful as a helper function in situations where there are too many papers in the corpus to hold in RAM at once.
    Given a filename for a PLOS XML paper,
    assembles a list of the papers it cites,
    collects DOIs, the number of mentions, and MICCs for each of those papers,
    Returns a dictionary of these measures, along with bare number of citations, keyed by DOI.
    Papers without discoverable DOIs are removed from the database.
    '''
    citations = {}

    paper = soupify(filename)

    # if verbose:
    paper_doi = plos_paper_doi(paper)
    print("DOI of paper is " + paper_doi)
    # Get the DOIs, the intra-paper mention counts, and the miccs.
    ipms = ipm_dictionary(paper)
    miccs = micc_dictionary(paper)
    if verbose:
        print("Retrieving DOIs for paper references...")
    dois = doi_batch(paper)
    # Remove the papers with un-identifiable DOIs.
    dois = {k: v for k, v in compress(list(dois.items()), list(dois.values()))}
    if verbose:
        print("Retrieved " + str(len(dois)) + " DOIs.")
        print("Processing database entries...")
    for j, doi in dois.items():
        citations[doi] = {}
        citations[doi]["ipms"] = ipms[j]
        citations[doi]["miccs"] = miccs[j]
        citations[doi]["citations"] = 1

    return citations


def large_citation_database(dois, xmlfolder="papers/", verbose=True, num_of_processors=8):
    '''
    Does the same thing as citation_database, but doesn't store as much information in RAM.
    Also, takes a list of DOIs as its argument, rather than a list of soupified XML papers.
    '''
    if verbose:
        print("Retrieving papers...")
    # The one-liner below does several things:
    # it retrieves the full XML text of the PLOS papers with the DOIs listed in the argument;
    # it saves the XML files to the path specified in xmlfolder, with filenames of the pattern <the part of the doi that comes after the slash>.xml;
    # it creates a list of those filenames.
    filenames = [remote_retrieve(doi, filename=xmlfolder + re.search(r"/.+", doi).group()[1:].encode("UTF-8") + ".xml") for doi in dois]

    # Prepare for multiprocessing!
    p = Pool(num_of_processors)

    # Do it!
    if verbose:
        print("Pulling citations...")
    individual_databases = p.map(paper_citations, filenames)
    # individual_databases = map(paper_citations, filenames)
    p.close()
    if verbose:
        print("Assembling citations into a database...")
    database = {}
    for db in individual_databases:
        for doi, data in db.items():
            try:
                database[doi]["ipms"].append(data["ipms"])
                database[doi]["miccs"].append(data["miccs"])
                database[doi]["citations"] += 1
            except KeyError:
                database[doi] = {}
                database[doi]["ipms"] = [data["ipms"]]
                database[doi]["miccs"] = [data["miccs"]]
                database[doi]["citations"] = 1
    if verbose:
        print("Database post-processing...")
    for info in database.values():
        info["median_ipm"] = median(info["ipms"])
        info["median_micc"] = median(info["miccs"])

    return database


def plos_search(query, query_type=None, rows=20, more_parameters=None,
                fq='''doc_type:full AND article_type_facet:"Research Article"''',
                output="json", verbose=False):
    '''
    Accesses the PLOS search API.
    query: the text of your query.
    query_type: subject, author, etc.
    rows: maximum number of results to return.
    more_parameters: an optional dictionary; key-value pairs are parameter names and values for the search api.
    fq: determines what kind of results are returned.
    Set by default to return only full documents that are research articles (almost always what you want).
    output: determines output type. Set to JSON by default, XML is also possible, along with a few others.
    Note that the PLOS search API cannot return full paper text, only abstracts and metadata.
    However, the URL for retrieving full-text PLOS papers is wholly specified by a paper's DOI (journal is not necessary).
    '''

    # This is *MY* API key. Please don't overuse it.
    # We're going to need a different one (or a different arrangement altogether) for any public release.
    api_key = "s4ZVBmgJyfZPMpqyy3Gs"  # for abecker@plos.org

    query_string = ""
    if query_type:
        query_string += query_type + ":"
    query_string += '"' + query + '"'

    params_string = ""
    if more_parameters:
        params_string = "&" + "&".join([key + "=" + quote(value) for key, value in more_parameters.items()])

    fq_string = "&fq=" + quote(fq)

    url = "https://api.plos.org/search?q=" + query_string + params_string + fq_string + "&wt=" + output + "&rows=" + str(rows) + "&api_key=" + api_key
    headers = {'Content-Type': 'application/' + output}
    if verbose:
        print(url)
    r = requests.get(url, headers=headers)
    r.encoding = "UTF-8"  # just to be sure
    return r.json()["response"]["docs"]


def plos_dois(search_results):
    '''Turns search results from plos_search into a list of DOIs.'''
    return [paper["id"] for paper in search_results]


def plos_paper_doi(paper):
    '''Given a soupified PLOS XML paper, returns that paper's DOI.'''
    paper_doi = paper.find("article-id", attrs={"pub-id-type": "doi"}).text
    return paper_doi


def zero_mentions(paper):
    '''
    Given a soupified PLOS XML paper, returns False,
    unless there are citation reference numbers in that paper's list of references that are not actually mentioned in the text of the paper.
    (This is against PLOS editorial policy, but it happens.)
    If there are such "zero-mention" citations, this function returns a tuple with two entries:
    the DOI of the given paper,
    and a list of all the zero-mention citation reference numbers.
    '''
    # Get the intra-paper mentions for everything in the list of references of the given paper.
    ipms = ipm_dictionary(paper)
    # See whether there are any references with zero mentions.
    if list(ipms.values()) and min(ipms.values()) > 0:
        print("Paper is okay!")
        return False
    else:
        print("Paper is NOT okay!")
        # Get the DOI of the offending paper
        doi = plos_paper_doi(paper)
        print(doi)
        # Get the reference numbers of the zero-mention references.
        zero_mentions = [x for x in list(ipms.items()) if not x[1]]
        print(zero_mentions)
        zero_mentions = [z[0] for z in zero_mentions]
        return (doi, zero_mentions)
